"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
import pytest

from flex.data.fed_dataset import Dataset
from flex.data.preprocessing_utils import normalize, one_hot_encoding


@pytest.fixture(name="fcd_ones")
def fixture_simple_fex_data_object_with_ones():
    X_data = np.ones(shape=(20, 5))
    y_data = np.random.choice(2, 20)
    return Dataset.from_array(X_data, y_data)


@pytest.fixture(name="fcd_zeros")
def fixture_simple_fex_data_object_with_zeros():
    X_data = np.zeros(shape=(20, 5))
    y_data = np.random.choice(2, 20)
    return Dataset.from_array(X_data, y_data)


def test_normalize_function(fcd_ones):
    X_data_normalized = normalize(fcd_ones).X_data.to_numpy()
    assert all(
        np.isclose(
            np.linalg.norm(X_data_normalized, axis=0),
            np.ones(X_data_normalized.shape[1]),
        )
    )
    assert fcd_ones.X_data.to_numpy().shape == X_data_normalized.shape


def test_normalize_func_norm_zero(fcd_zeros):
    X_data_normalized = normalize(fcd_zeros).X_data.to_numpy()
    assert np.array_equal(fcd_zeros.X_data.to_numpy(), X_data_normalized)
    assert fcd_zeros.X_data.to_numpy().shape == X_data_normalized.shape


def test_one_hot_encoding(fcd_ones):
    new_fcd = one_hot_encoding(fcd_ones, n_labels=2)
    assert new_fcd.y_data.to_numpy().shape[1] == 2
    assert len(new_fcd.y_data) == len(fcd_ones.y_data)


def test_one_hot_encoding_error(fcd_zeros):
    with pytest.raises(ValueError):
        one_hot_encoding(fcd_zeros)



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from torch.utils.data import Dataset
from torchvision.datasets import VisionDataset


class DefaultVision(VisionDataset):
    def __init__(self, data, transform=None, target_transform=None):
        super().__init__(
            root="",
            transforms=None,
            transform=transform,
            target_transform=target_transform,
        )
        self.data = data

    def __getitem__(self, index: int):
        image, label = self.data[index]
        if self.transform:
            image = self.transform(image)
        if self.target_transform:
            label = self.target_transform(label)
        return image, label

    def __len__(self):
        return len(self.data)


class FeatureDataset(Dataset):
    def __init__(self, dataset: Dataset):
        self._dataset = dataset

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, idx):
        return self._dataset[idx][0]


class LabelDataset(Dataset):
    def __init__(self, dataset: Dataset):
        self._dataset = dataset

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, idx):
        return self._dataset[idx][1]



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import copy
import itertools
import warnings
from collections import deque
from typing import Iterable, Union

import numpy as np


class LazyIndexable:
    def __init__(
        self,
        iterable: Iterable,
        length: int,
        iterable_indexes=None,
        storage: dict = None,
        last_access: deque = None,
    ):
        if iterable_indexes is None:
            iterable_indexes = np.arange(length, dtype=np.uint32)
        if storage is None:
            storage = {}
        self._iterable = iterable
        self._len = length
        self._iterable_indexes = np.asarray(iterable_indexes, dtype=np.uint32)
        self._storage = storage
        if last_access is None:
            self._last_access = deque(maxlen=1)
            self._last_access.append(0)
        else:
            self._last_access = last_access
        try:
            self._iterable[0]
            self._is_generator = False
        except TypeError:  # object is not subscriptable
            self._is_generator = True

    def __repr__(self) -> str:
        return (
            f"len:{self._len}\n"
            f"is_generator:{self._is_generator}\n"
            f"iterable:{self._iterable}\n"
            f"iterable_indexes:{self._iterable_indexes}\n"
            f"storage:{self._storage}"
        )

    def __len__(self) -> int:
        return self._len

    def __getitem_with_seq(self, s: Union[slice, list, np.ndarray]):
        return LazyIndexable(
            self._iterable,
            iterable_indexes=self._iterable_indexes[s],
            length=len(self._iterable_indexes[s]),
            storage=self._storage,
            last_access=self._last_access,
        )

    def __getitem_with_int(self, s):
        index = self._iterable_indexes[s]
        if not self._is_generator:
            return self._iterable[index]
        if index in self._storage:
            return self._storage[index]
        start = self._last_access[-1]
        # If it is not in the storage, we must consume the iterable
        for i, element in enumerate(self._iterable, start=start):
            self._storage[i] = element  # Every we consume is stored for later usage
            self._last_access.append(i + 1)
            if i == index:
                return element
        # Value not found in consumable
        raise IndexError(f"Index {index} out of range")

    def __getitem__(self, s: Union[int, slice, list]):
        #  Proceed with the actual getitem logic
        try:
            s = int(s)
            val = self.__getitem_with_int(s)
        except TypeError:
            val = self.__getitem_with_seq(s)
        return val

    def tolist(self):
        """Function that returns the LazyIndexable as list."""
        return list(self)

    def to_numpy(self, dtype=None):
        if isinstance(self._iterable, (np.ndarray, np.generic)):
            return self._iterable[self._iterable_indexes]
        return np.array(self.tolist(), dtype=dtype)

    def __getstate__(self):
        """Required to make LazyIndexable pickable when self._is_generator==True."""
        state_dict = copy.copy(self.__dict__)
        iterable_is_consumed = False
        for key in state_dict:
            if key == "_iterable" and state_dict["_is_generator"]:
                warnings.warn(  # noqa: B028
                    "Pickling an LazyIndexable fully loads its into memory",
                    RuntimeWarning,
                )
                state_dict[key] = self.tolist()
                iterable_is_consumed = True
        if iterable_is_consumed:  # Ditch is_generator state and _storage
            state_dict["_is_generator"] = False
            state_dict["_storage"] = {}
        return state_dict

    def __deepcopy__(self, memo):
        # """Overwrites deepcopy method."""
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == "_iterable" and self._is_generator:
                self._iterable, new_iterable = itertools.tee(self._iterable, 2)
                setattr(result, k, new_iterable)
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        return result



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import functools
from typing import List

from flex.common.utils import check_min_arguments
from flex.model import FlexModel


def ERROR_MSG_MIN_ARG_GENERATOR(f, min_args):
    return f"The decorated function: {f.__name__} is expected to have at least {min_args} argument/s."


def init_server_model(func):
    @functools.wraps(func)
    def _init_server_model_(server_flex_model: FlexModel, _, *args, **kwargs):
        server_flex_model.update(func(*args, **kwargs))

    return _init_server_model_


def deploy_server_model(func):
    min_args = 1
    assert check_min_arguments(func, min_args), ERROR_MSG_MIN_ARG_GENERATOR(
        func, min_args
    )

    @functools.wraps(func)
    def _deploy_model_(
        server_flex_model: FlexModel,
        clients_flex_models: List[FlexModel],
        *args,
        **kwargs,
    ):
        for k in clients_flex_models:
            # Reminder, it is not possible to make assignements here
            clients_flex_models[k].update(func(server_flex_model, *args, **kwargs))

    return _deploy_model_


def collect_clients_weights(func):
    min_args = 1
    assert check_min_arguments(func, min_args), ERROR_MSG_MIN_ARG_GENERATOR(
        func, min_args
    )

    @functools.wraps(func)
    def _collect_weights_(
        aggregator_flex_model: FlexModel,
        clients_flex_models: List[FlexModel],
        *args,
        **kwargs,
    ):
        if "weights" not in aggregator_flex_model:
            aggregator_flex_model["weights"] = []
        for k in clients_flex_models:
            client_weights = func(clients_flex_models[k], *args, **kwargs)
            aggregator_flex_model["weights"].append(client_weights)

    return _collect_weights_


def aggregate_weights(func):
    min_args = 1
    assert check_min_arguments(func, min_args), ERROR_MSG_MIN_ARG_GENERATOR(
        func, min_args
    )

    @functools.wraps(func)
    def _aggregate_weights_(aggregator_flex_model: FlexModel, _, *args, **kwargs):
        aggregator_flex_model["aggregated_weights"] = func(
            aggregator_flex_model["weights"], *args, **kwargs
        )
        aggregator_flex_model["weights"] = []

    return _aggregate_weights_


def set_aggregated_weights(func):
    min_args = 2
    assert check_min_arguments(func, min_args), ERROR_MSG_MIN_ARG_GENERATOR(
        func, min_args
    )

    @functools.wraps(func)
    def _deploy_aggregated_weights_(
        aggregator_flex_model: FlexModel,
        servers_flex_models: FlexModel,
        *args,
        **kwargs,
    ):
        for k in servers_flex_models:
            func(
                servers_flex_models[k],
                aggregator_flex_model["aggregated_weights"],
                *args,
                **kwargs,
            )

    return _deploy_aggregated_weights_


def evaluate_server_model(func):
    min_args = 1
    assert check_min_arguments(func, min_args), ERROR_MSG_MIN_ARG_GENERATOR(
        func, min_args
    )

    @functools.wraps(func)
    def _evaluate_server_model_(server_flex_model: FlexModel, _, *args, **kwargs):
        return func(server_flex_model, *args, **kwargs)

    return _evaluate_server_model_



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from dataclasses import asdict, dataclass
from typing import Hashable, List, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt


class InvalidConfig(ValueError):
    """Raised when the input config is wrong"""

    pass


@dataclass
class FedDatasetConfig:
    """Class used to represent a configuration to federate a centralized dataset.
    The following table shows the compatiblity of each option:

    +-------------------------+-------------+--------------+-------------+-----------------------+-----------------+---------------------+-----------------------+----------------------+--------------------------+-----------------+-------------+
    | Options compatibility   | **n_nodes** | **node_ids** | **weights** | **weights_per_label** | **replacement** | **labels_per_node** | **features_per_node** | **indexes_per_node** | **group_by_label_index** | **keep_labels** | **shuffle** |
    +=========================+=============+==============+=============+=======================+=================+=====================+=======================+======================+==========================+=================+=============+
    | **n_nodes**             | -           | Y            | Y           | Y                     | Y               | Y                   | Y                     | N                    | N                        | Y               | Y           |
    +-------------------------+-------------+--------------+-------------+-----------------------+-----------------+---------------------+-----------------------+----------------------+--------------------------+-----------------+-------------+
    | **node_ids**            | -           | -            | Y           | Y                     | Y               | Y                   | Y                     | Y                    | N                        | Y               | Y           |
    +-------------------------+-------------+--------------+-------------+-----------------------+-----------------+---------------------+-----------------------+----------------------+--------------------------+-----------------+-------------+
    | **weights**             | -           | -            | -           | N                     | Y               | Y                   | Y                     | N                    | N                        | Y               | Y           |
    +-------------------------+-------------+--------------+-------------+-----------------------+-----------------+---------------------+-----------------------+----------------------+--------------------------+-----------------+-------------+
    | **weights_per_label**   | -           | -            | -           | -                     | Y               | N                   | N                     | N                    | N                        | Y               | Y           |
    +-------------------------+-------------+--------------+-------------+-----------------------+-----------------+---------------------+-----------------------+----------------------+--------------------------+-----------------+-------------+
    | **replacement**         | -           | -            | -           | -                     | -               | Y                   | N                     | N                    | N                        | Y               | Y           |
    +-------------------------+-------------+--------------+-------------+-----------------------+-----------------+---------------------+-----------------------+----------------------+--------------------------+-----------------+-------------+
    | **labels_per_node**     | -           | -            | -           | -                     | -               | -                   | N                     | N                    | N                        | Y               | Y           |
    +-------------------------+-------------+--------------+-------------+-----------------------+-----------------+---------------------+-----------------------+----------------------+--------------------------+-----------------+-------------+
    | **features_per_node**   | -           | -            | -           | -                     | -               | -                   | -                     | N                    | N                        | Y               | Y           |
    +-------------------------+-------------+--------------+-------------+-----------------------+-----------------+---------------------+-----------------------+----------------------+--------------------------+-----------------+-------------+
    | **indexes_per_node**    | -           | -            | -           | -                     | -               | -                   | -                     | -                    | N                        | Y               | Y           |
    +-------------------------+-------------+--------------+-------------+-----------------------+-----------------+---------------------+-----------------------+----------------------+--------------------------+-----------------+-------------+
    | **group_by_label_index**| -           | -            | -           | -                     | -               | -                   | -                     | -                    | -                        | N               | Y           |
    +-------------------------+-------------+--------------+-------------+-----------------------+-----------------+---------------------+-----------------------+----------------------+--------------------------+-----------------+-------------+
    | **keep_labels**         | -           | -            | -           | -                     | -               | -                   | -                     | -                    | -                        | -               | Y           |
    +-------------------------+-------------+--------------+-------------+-----------------------+-----------------+---------------------+-----------------------+----------------------+--------------------------+-----------------+-------------+
    | **shuffle**             | -           | -            | -           | -                     | -               | -                   | -                     | -                    | -                        | -               | -           |
    +-------------------------+-------------+--------------+-------------+-----------------------+-----------------+---------------------+-----------------------+----------------------+--------------------------+-----------------+-------------+

    Attributes
    ----------
    seed: Optional[int]
        Seed used to make the federated dataset generated reproducible with this configuration. Default None.
    n_nodes: int
        Number of nodes among which to split a centralized dataset. Default 2.
    shuffle: bool
        If True data is shuffled before being sampled. Default False.
    node_ids: Optional[List[Hashable]]
        Ids to identifty each node, if not provided, nodes will be indexed using integers. If n_nodes is also \
        given, we consider up to n_nodes elements. Default None.
    weights: Optional[npt.NDArray]
        A numpy.array which provides the proportion of data to give to each node. Default None.
    weights_per_label: Optional[npt.NDArray]
        A numpy.array which provides the proportion of data to give to each node and class of the dataset to federate. \
        We expect a bidimensional array of shape (n, m) where "n" is the number of nodes and "m" is the number of labels of \
        the dataset to federate. Default None.
    replacement: bool
        Whether the samping procedure used to split a centralized dataset is with replacement or not. Default False
    labels_per_node: Optional[Union[int, npt.NDArray, Tuple[int]]]
        labels to assign to each node, if provided as an int, it is the number labels per node, if provided as a \
        tuple of ints, it establishes a mininum and a maximum of number of labels per node, a random number sampled \
        in such interval decides the number of labels of each node. If provided as a list of lists, it establishes the labels \
        assigned to each node. Default None.
    features_per_node: Optional[Union[int, npt.NDArray, Tuple[int]]]
        Features to assign to each node, it share the same interface as labels_per_node. Default None.
    indexes_per_node: Optional[npt.NDArray]
        Data indexes to assign to each node. Default None.
    group_by_label_index: Optional[int]
        Index which indicates which feature unique values will be used to generate federated nodes. Default None.
    keep_labels: Optional[list[bool]]
        Whether each node keeps or not the labels or y_data
    """

    seed: Optional[int] = None
    n_nodes: int = 2
    shuffle: bool = False
    node_ids: Optional[List[Hashable]] = None
    weights: Optional[npt.NDArray] = None
    weights_per_label: Optional[npt.NDArray] = None
    replacement: bool = False
    labels_per_node: Optional[Union[int, npt.NDArray, Tuple[int]]] = None
    features_per_node: Optional[Union[int, npt.NDArray, Tuple[int]]] = None
    indexes_per_node: Optional[npt.NDArray] = None
    group_by_label_index: Optional[int] = None
    keep_labels: Optional[List[bool]] = None

    def _check_incomp(self, dict, option1, option2):
        """This function checks if two options are compatible, if not it raises and exception"""
        cond1 = dict[option1] is not None if option1 != "replacement" else dict[option1]
        cond2 = dict[option2] is not None if option2 != "replacement" else dict[option2]
        if cond1 and cond2:
            raise InvalidConfig(
                f"Options {option1} and {option2} are incompatible, please provide only one."
            )

    def validate(self):
        """This function checks whether the configuration to federate a dataset is correct."""
        self_dict = asdict(self)
        # By default every option is compatible, therefore we only specify incompatibilities
        self._check_incomp(self_dict, "weights", "group_by_label_index")
        self._check_incomp(self_dict, "weights", "weights_per_label")
        self._check_incomp(self_dict, "weights", "indexes_per_node")
        self._check_incomp(self_dict, "weights", "group_by_label_index")
        self._check_incomp(self_dict, "weights_per_label", "indexes_per_node")
        self._check_incomp(self_dict, "weights_per_label", "labels_per_node")
        self._check_incomp(self_dict, "weights_per_label", "features_per_node")
        self._check_incomp(self_dict, "weights_per_label", "indexes_per_node")
        self._check_incomp(self_dict, "weights_per_label", "group_by_label_index")
        self._check_incomp(self_dict, "replacement", "indexes_per_node")
        self._check_incomp(self_dict, "replacement", "group_by_label_index")
        self._check_incomp(self_dict, "labels_per_node", "features_per_node")
        self._check_incomp(self_dict, "labels_per_node", "indexes_per_node")
        self._check_incomp(self_dict, "labels_per_node", "group_by_label_index")
        self._check_incomp(self_dict, "features_per_node", "indexes_per_node")
        self._check_incomp(self_dict, "features_per_node", "group_by_label_index")

        self.__validate_nodes_and_weights()
        if self.indexes_per_node is not None:
            self.__validate_indexes_per_node()
        elif self.labels_per_node is not None:
            self.__validate_labels_per_node()
        elif self.features_per_node is not None:
            self.__validate_features_per_node()
        if self.keep_labels is not None:
            self.__validate_keep_labels()

    def __validate_indexes_per_node(self):
        if len(self.indexes_per_node) != self.n_nodes:
            raise InvalidConfig(
                "The number of provided nodes should equal the length of indexes per node."
            )

    def __validate_keep_labels(self):
        if len(self.keep_labels) != self.n_nodes:
            raise InvalidConfig(
                "keep_labels list should have the same length as n_nodes."
            )

    def __validate_nodes_and_weights(self):
        if self.n_nodes < 2:
            raise InvalidConfig(
                "The number of nodes must be greater or equal to 2. Default is 2"
            )

        if self.node_ids is not None and self.n_nodes > len(self.node_ids):
            raise InvalidConfig(
                "The number of named nodes, node_ids, can not be greater than the number of nodes, n_nodes"
            )

        if self.weights is not None and self.n_nodes != len(self.weights):
            raise InvalidConfig("The number of weights must equal the number of nodes.")

        if (
            self.weights_per_label is not None
            and len(np.asarray(self.weights_per_label).shape) != 2
        ):
            raise InvalidConfig(
                (
                    "weights_per_label must be a two dimensional array where the first dimension is the number of nodes and the second is the number of labels of the dataset to federate."
                )
            )
        if self.weights_per_label is not None and self.n_nodes != len(
            self.weights_per_label
        ):
            raise InvalidConfig(
                "The length of weights_per_label must equal the number of nodes."
            )

        if self.weights is not None and max(self.weights) > 1:
            raise InvalidConfig(
                "Provided weights contains an element greater than 1, we do not allow sampling more than one time the entire dataset per node."
            )
        if self.weights is not None and min(self.weights) < 0:
            raise InvalidConfig(
                "Provided weights contains negative numbers, we do not allow that."
            )

    def __validate_labels_per_node(self):
        if isinstance(self.labels_per_node, tuple):
            if len(self.labels_per_node) != 2:
                raise InvalidConfig(
                    f"labels_per_node if provided as a tuple, it must have two elements, mininum number of labels per node and maximum number of labels per node, but labels_per_node={self.labels_per_node}."
                )
        elif not isinstance(self.labels_per_node, int) and self.n_nodes != len(
            self.labels_per_node
        ):
            raise InvalidConfig(
                "labels_per_node if provided as a list o np.ndarray, its length and n_nodes must equal."
            )

    def __validate_features_per_node(self):
        if not self.replacement:
            raise InvalidConfig(
                "By setting replacement to False and specifying features_per_node, nodes will not share any data instances."
            )
        if isinstance(self.features_per_node, tuple):
            if len(self.features_per_node) != 2:
                raise InvalidConfig(
                    f"features_per_node if provided as a tuple, it must have two elements, mininum number of features per node and maximum number of features per node, but features_per_node={self.features_per_node}."
                )
        elif not isinstance(self.features_per_node, int) and self.n_nodes != len(
            self.features_per_node
        ):
            raise InvalidConfig(
                "features_per_node if provided as a list o np.ndarray, its length and n_nodes must equal."
            )



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import List

import numpy as np

from flex.distributed.proto.tensor_pb2 import Tensor


def toTensorList(tensors: List[np.ndarray]) -> List[Tensor]:
    rv = []
    for tensor in tensors:
        rv.append(
            Tensor(
                shape=list(tensor.shape), data=tensor.tobytes(), dtype=tensor.dtype.name
            )
        )
    return rv


def toNumpyArray(tensors: List[Tensor]) -> List[np.ndarray]:
    rv = []
    for tensor in tensors:
        shape = tuple(tensor.shape)
        dtype = tensor.dtype if tensor.dtype.strip() != "" else "float32"
        t = np.frombuffer(tensor.data, dtype=dtype).reshape(shape)
        rv.append(t)
    return rv



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

import functools
import random
from typing import Callable, Hashable, Union

from flex.actors.actors import FlexActors, FlexRoleManager
from flex.actors.architectures import client_server_architecture, p2p_architecture
from flex.data import FedDataset
from flex.model.model import FlexModel


class FlexPool:
    """
    Class that orchest the training phase of a federated learning experiment.
    The FlexPool class is responsible for orchestating the nodes to train a
    federated model.
    This class represents a pool of actors and is in charge of checking the
    communications between them during the process of training a federated model.

    Note: At the moment this class only supports Horizontal Federated Learning,
    but in the future it will cover Vertical Federated Learning and Transfer Learning,
    so users can simulate all the experiments correctly.

    Attributes
    ----------
        - flex_data (FedDataset): The federated dataset prepared to be used.
        - flex_actors (FlexActors): Actors with its roles.
        - flex_models (defaultdict): A dictionary containing the each actor id,
        and initialized to None. The model to train by each actor will be initialized
        using the map function following the communication constraints.


    We offer two class methods to create two architectures, client-server architecture
    and p2p architecture. In the client-server architecture, every id from the
    FedDataset is assigned to be a client, and we create a third-party actor,
    supposed to be neutral, to orchestate the training. Meanwhile, in the p2p
    architecture, each id from the FedDataset will be assigned to be client,
    server and aggregator. In both cases, the method will create the actors
    so the user will only have to apply the map function to train the model.

    If the user wants to use a different architecture, she will need to create
    the actors by using the FlexActors class. For example, we let the user create
    a client-server architecture with multiple aggregators, to carry out the aggregation.
    """

    def __init__(
        self,
        flex_data: FedDataset,
        flex_actors: FlexActors,
        flex_models: dict[Hashable, FlexModel] = None,
    ) -> None:
        self._actors = flex_actors  # Actors
        self._data = flex_data  # FedDataset
        self._models = flex_models
        if self._models is None:
            self._models = {k: FlexModel() for k in self._actors}
            for k in self._models:
                self._models[k].actor_id = k
        self.validate()  # check if the provided arguments generate a valid object

    @classmethod
    def check_compatibility(cls, src_pool, dst_pool):
        """Method to check the compatibility between two different pools.
        This method is used by the map function to check if the
        function to apply from the source pool to the destination pool can be done.

        Args:
            src_pool (FlexPool): Source pool. Pool that will send the message.
            dst_pool (FlexPool): Destination pool. Pool that will recieve the message.

        Returns:
            bool: True if pools are compatible. False in other case.
        """
        return all(
            FlexRoleManager.check_compatibility(src, dst)
            for _, src in src_pool._actors.items()
            for _, dst in dst_pool._actors.items()
        )

    def map(self, func: Callable, dst_pool: FlexPool = None, **kwargs):
        r"""Method used to send messages from one pool to another. The pool using
        this method is the source pool, and it will send a message, apply a function,
        to the destination pool. If no destination pool is provided, then the function is applied
        to the source pool. The pool sends a message in order to complete a round
        in the Federated Learning (FL) paradigm, so, some examples of the messages
        that will be used by the pools are:
        - send_model: Aggregators send the model to the server when the aggregation is done.
        - aggregation_step: Clients send the model to the aggregator so it can apply the
        aggregation mechanism given.
        - deploy_model: Server sends the global model to the clients once the weights has
        been aggregated.
        - init_model: Server sends the model to train during the learning phase, so the
        clients can initialize it. This is a particular case from the deploy_model case.

        Args:
        -----
            func (Callable): If dst_pool is None, then message is sent to the source (self). In this situation
            the function func is called for each actor in the pool, providing actor's data and actor's model
            as arguments in addition to \*args and \**kwargs. If dst_pool is not None, the message is sent from
            the source pool (self) to the destination pool (dst_pool). The function func is called for each actor
            in the pool, providing the model of the current actor in the source pool and all the models of the
            actors in the destination pool.

            dst_pool (FlexPool): Pool that will recieve the message from the source pool (self), it can be None.

        Raises:
        -------
            ValueError: This method raises and error if the pools aren't allowed to comunicate

        Returns:
        --------
            List[Any]: A list of the result of applying the function (func) from the source pool (self) to the
            destination pool (dst_pool). If dst_pool is None, then the results come from the source pool. The
            length of the returned values equals the number of actors in the source pool.
        """
        if dst_pool is None:
            res = [
                func(self._models.get(i), self._data.get(i), **kwargs)
                for i in self._actors
            ]
        elif FlexPool.check_compatibility(self, dst_pool):
            res = [
                func(self._models.get(i), dst_pool._models, **kwargs)
                for i in self._actors
            ]
        else:
            raise ValueError(
                "Source and destination pools are not allowed to comunicate, ensure that their actors can communicate."
            )
        if all(ele is not None for ele in res):
            return res

    def select(self, criteria: Union[int, Callable], *criteria_args, **criteria_kwargs):
        """Function that returns a subset of a FlexPool meeting a certain criteria.
        If criteria is an integer, a subset of the available nodes of size criteria is
        randomly sampled. If criteria is a function, then we select those nodes where
        the function returns True values. Note that, the function must have at least
        two arguments, a node id and the roles associated to such node id.
        The actor_id is a string, and the actor_role is a FlexRole object.

        Note: This function doesn't send a copy of the original pool, it sends a reference.
            Changes made on the returned pool affect the original pool.

        Args:
        -----
            criteria (int, Callable): if a function is provided, then it must return
            True/False values for each pair of node_id, node_roles. Additional arguments
            required for the function are passed in criteria_args and criteria_kwargs.
            Otherwise, criteria is interpreted as number of nodes to randomly sample from the pool.
            criteria_args: additional args required for the criteria function. Otherwise ignored.
            criteria_kwargs: additional keyword args required for the criteria function. Otherwise ignored.

        Returns:
        --------
            FlexPool: a pool that contains the nodes that meet the criteria.
        """
        new_actors = FlexActors()
        new_data = FedDataset()
        new_models = {}
        available_nodes = list(self._actors.keys())
        if callable(criteria):
            selected_keys = [
                actor_id
                for actor_id in available_nodes
                if criteria(
                    actor_id, self._actors[actor_id], *criteria_args, **criteria_kwargs
                )
            ]
        else:
            num_nodes = criteria
            selected_keys = random.sample(available_nodes, num_nodes)

        for actor_id in selected_keys:
            new_actors[actor_id] = self._actors[actor_id]
            new_models[actor_id] = self._models[actor_id]
            if actor_id in self._data:
                new_data[actor_id] = self._data[actor_id]

        return FlexPool(
            flex_actors=new_actors, flex_data=new_data, flex_models=new_models
        )

    def __len__(self):
        return len(self._actors)

    @functools.cached_property
    def actor_ids(self):
        return list(self._actors.keys())

    @functools.cached_property
    def clients(self):
        """Property to get all the clients available in a pool.

        Returns:
        --------
            FlexPool: Pool containing all the clients from a pool
        """
        return self.select(lambda a, b: FlexRoleManager.is_client(b))

    @functools.cached_property
    def aggregators(self):
        """Property to get all the aggregator available in a pool.

        Returns:
        --------
            FlexPool: Pool containing all the aggregators from a pool
        """
        return self.select(lambda a, b: FlexRoleManager.is_aggregator(b))

    @functools.cached_property
    def servers(self):
        """Property to get all the servers available in a pool.

        Returns:
        --------
            FlexPool: Pool containing all the servers from a pool
        """
        return self.select(lambda a, b: FlexRoleManager.is_server(b))

    def validate(self):  # sourcery skip: de-morgan
        """Function that checks whether the object is correct or not."""
        actors_ids = self._actors.keys()
        data_ids = self._data.keys()
        models_ids = self._models.keys()
        for actor_id in actors_ids:
            if (
                FlexRoleManager.is_client(self._actors[actor_id])
                and actor_id not in data_ids
            ):
                raise ValueError(
                    "All node with client role must have data. Node with client role and id {actor_id} does not have any data."
                )
            if actor_id not in self._models:
                raise ValueError(
                    f"All nodes must have a FlexModel object associated as a model, but {actor_id} does not."
                )
        flex_models_ids = {self._models[k].actor_id for k in self._models}
        if not (
            actors_ids >= data_ids
            and actors_ids >= models_ids
            and actors_ids >= flex_models_ids
        ):  # noqa: E501
            raise ValueError("Each node with data or model must have a role asigned")

    @classmethod
    def client_server_pool(
        cls,
        fed_dataset: FedDataset,
        init_func: Callable,
        server_id: str = "server",
        **kwargs,
    ):
        """Method to create a client-server architeture for a FlexDataset given.
        This functions is used when you have a FlexDataset and you want to start
        the learning phase following a traditional client-server architecture.

        This method will assing to each id from the FlexDataset the client-role,
        and will create a new actor that will be the server-aggregator that will
        orchestrate the learning phase.

        Args:
        -----
            fed_dataset (FedDataset): Federated dataset used to train a model.

        Returns:
        --------
            FlexPool: A FlexPool with the assigned roles for a client-server architecture.
        """
        actors = client_server_architecture(fed_dataset.keys(), server_id)

        new_arch = cls(
            flex_data=fed_dataset,
            flex_actors=actors,
            flex_models=None,
        )
        new_arch.servers.map(init_func, **kwargs)
        return new_arch

    @classmethod
    def p2p_pool(cls, fed_dataset: FedDataset, init_func: Callable, **kwargs):
        """Method to create a peer-to-peer (p2p) architecture for a FlexDataset given.
        This method is used when you have a FlexDataset and you want to start the
        learning phase following a p2p architecture.

        This method will assing all roles (client-aggregator-server) to every id from
        the FlexDataset, so each participant in the learning phase can act as client,
        aggregator and server.

        Args:
        -----
            fed_dataset (FedDataset): Federated dataset used to train a model.

        Returns:
        --------
            FlexPool: A FlexPool with the assigned roles for a p2p architecture.
        """
        new_arch = cls(
            flex_data=fed_dataset,
            flex_actors=p2p_architecture(fed_dataset),
            flex_models=None,
        )
        new_arch.servers.map(init_func, **kwargs)
        return new_arch

    def __iter__(self):
        yield from self._actors



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import unittest

import numpy as np
import pytest

from flex.data import FedDatasetConfig
from flex.data.fed_dataset_config import InvalidConfig


class TestFlexDatasetConfig(unittest.TestCase):
    def test_nclients(self):
        a = FedDatasetConfig(n_nodes=1)
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_client_names_error(self):
        a = FedDatasetConfig(node_ids=["Pepe"])
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_weights(self):
        a = FedDatasetConfig(weights=[0.5])
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_max_weights(self):
        a = FedDatasetConfig(weights=[1.2, 0.3])
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_positive_weights(self):
        a = FedDatasetConfig(weights=[-1.2, 0.3])
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_mutually_exclusive_options(self):
        a = FedDatasetConfig(labels_per_node=2, features_per_node=2)
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_labels_per_client_tuple(self):
        a = FedDatasetConfig(labels_per_node=(1, 2, 3))
        with pytest.raises(InvalidConfig):
            a.validate()
        a.labels_per_node = (1,)
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_labels_per_client_arr(self):
        a = FedDatasetConfig(labels_per_node=[0, [0, 2], 1])
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_features_per_client_arr(self):
        a = FedDatasetConfig(features_per_node=[[0, 1], [2, 3], [4]], replacement=True)
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_features_per_client_tuple(self):
        a = FedDatasetConfig(features_per_node=(1, 2, 3), replacement=True)
        with pytest.raises(InvalidConfig):
            a.validate()
        a.features_per_node = (1,)
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_features_per_client_w_replacement(self):
        a = FedDatasetConfig(replacement=False, features_per_node=3)
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_incompatible_options_w_indexes_per_client(self):
        a = FedDatasetConfig(n_nodes=3, features_per_node=3, indexes_per_node=[[2]])
        b = FedDatasetConfig(n_nodes=3, labels_per_node=3, indexes_per_node=[[2]])
        FedDatasetConfig(n_nodes=3, keep_labels=[True] * 3, indexes_per_node=[[2]])
        with pytest.raises(InvalidConfig):
            a.validate()
        with pytest.raises(InvalidConfig):
            b.validate()

    def test_indexes_per_client_w_number_of_clients(self):
        a = FedDatasetConfig(replacement=False, indexes_per_node=[])
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_indexes_per_client_w_replacement(self):
        a = FedDatasetConfig(replacement=True, indexes_per_node=[])
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_weights_incompatibility(self):
        a = FedDatasetConfig(weights=[1, 2], weights_per_label=np.ones((2, 2)))
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_weights_per_class_incompatibility(self):
        a = FedDatasetConfig(labels_per_node=2, weights_per_label=np.ones((2, 2)))
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_weights_per_class_shape(self):
        a = FedDatasetConfig(weights_per_label=np.ones((1,)))
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_weights_per_class_size(self):
        a = FedDatasetConfig(n_nodes=3, weights_per_label=np.ones((1, 2)))
        with pytest.raises(InvalidConfig):
            a.validate()

    def test_keep_labels(self):
        a = FedDatasetConfig(n_nodes=3, keep_labels=[True])
        with pytest.raises(InvalidConfig):
            a.validate()



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import unittest

import numpy as np
import pytest

from flex.data.dataset import Dataset
from flex.data.fed_dataset import FedDataset
from flex.data.preprocessing_utils import normalize


@pytest.fixture(name="fld")
def fixture_flex_dataset():
    """Function that returns a FlexDataset provided as example to test functions.

    Returns:
        FedDataset: A FlexDataset generated randomly
    """
    X_data = np.random.rand(100).reshape([20, 5])
    y_data = np.random.choice(2, 20)
    fcd = Dataset.from_array(X_data, y_data)
    X_data = np.random.rand(100).reshape([20, 5])
    y_data = np.random.choice(2, 20)
    fcd1 = Dataset.from_array(X_data, y_data)
    X_data = np.random.rand(100).reshape([20, 5])
    y_data = np.random.choice(2, 20)
    fcd2 = Dataset.from_array(X_data, y_data)
    return FedDataset({"client_1": fcd, "client_2": fcd1, "client_3": fcd2})


@pytest.fixture(name="fcd")
def fixture_simple_fex_data_object():
    X_data = np.random.rand(100).reshape([20, 5])
    y_data = np.random.choice(2, 20)
    return Dataset.from_array(X_data, y_data)


class TestFlexDataset(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def _fixture_flex_dataset(self, fld):
        self._fld = fld

    def test_get_method(self):
        X_data = np.random.rand(100).reshape([20, 5])
        y_data = np.random.choice(2, 20)
        fcd = Dataset.from_array(X_data, y_data)
        flex_data = FedDataset()
        flex_data["client_1"] = fcd
        assert flex_data["client_1"] == fcd
        assert flex_data.get("client_2") is None

    def test_normalize_method(self):
        new_fld = self._fld.normalize()
        assert all(
            not np.array_equal(
                client_orig.X_data.to_numpy(), client_mod.X_data.to_numpy()
            )
            for client_orig, client_mod in zip(self._fld.values(), new_fld.values())
        )

    def test_one_hot_encoding(self):
        new_fld = self._fld.one_hot_encoding(n_labels=2)
        assert all(
            client.y_data.to_numpy().shape[1] == 2 for _, client in new_fld.items()
        )

    def test_map_method(self):
        new_fld = self._fld.apply(func=normalize)
        assert all(
            not np.array_equal(client_orig.X_data, client_mod.X_data)
            for client_orig, client_mod in zip(self._fld.values(), new_fld.values())
        )

    def test_map_func_from_outside(self):
        def dummy_func(data, **kwargs):
            return data

        new_fld = self._fld.apply(func=dummy_func, num_proc=2)
        assert all(
            np.array_equal(client_orig.X_data, client_mod.X_data)
            for client_orig, client_mod in zip(self._fld.values(), new_fld.values())
        )

    def test_proprocessing_custom_func_more_processes_than_clients(self):
        new_fld = self._fld.apply(func=normalize, num_proc=10)
        assert all(
            not np.array_equal(client_orig.X_data, client_mod.X_data)
            for client_orig, client_mod in zip(self._fld.values(), new_fld.values())
        )

    def test_chosen_clients_custom_func(self):
        chosen_clients = ["client_1", "client_2"]
        new_fld = self._fld.apply(func=normalize, num_proc=10, node_ids=chosen_clients)
        assert any(
            np.array_equal(client_orig.X_data, client_mod.X_data)
            for client_orig, client_mod in zip(self._fld.values(), new_fld.values())
        )

    def test_chosen_clients_normalize_data_func(self):
        chosen_clients = ["client_1", "client_2"]
        new_fld = self._fld.normalize(num_proc=10, node_ids=chosen_clients)
        assert any(
            np.array_equal(client_orig.X_data, client_mod.X_data)
            for client_orig, client_mod in zip(self._fld.values(), new_fld.values())
        )

    def test_all_clients_in_flex_dataset_when_mapping_func(self):
        client_ids = ["client_1", "client_84"]
        with pytest.raises(ValueError):
            self._fld.apply(func=normalize, num_proc=10, node_ids=client_ids)

    def test_map_func_executes_secuential(self):
        chosen_clients = ["client_1"]
        new_fld = self._fld.normalize(num_proc=1, node_ids=chosen_clients)
        assert any(
            np.array_equal(client_orig.X_data, client_mod.X_data)
            for client_orig, client_mod in zip(self._fld.values(), new_fld.values())
        )

    def test_map_recieves_one_client_as_str_correct(self):
        chosen_clients = "client_1"
        new_fld = self._fld.normalize(num_proc=1, node_ids=chosen_clients)
        assert any(
            np.array_equal(client_orig.X_data, client_mod.X_data)
            for client_orig, client_mod in zip(self._fld.values(), new_fld.values())
        )

    def test_map_recieves_one_client_as_str_fails(self):
        client_ids = "client_8232"
        with pytest.raises(ValueError):
            self._fld.apply(func=normalize, num_proc=1, node_ids=client_ids)



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import unittest
from collections import defaultdict

import numpy as np
import pytest
from sklearn.datasets import load_iris

from flex.actors.actors import FlexActors, FlexRole, FlexRoleManager
from flex.data import FedDataDistribution
from flex.data.fed_dataset import Dataset, FedDataset
from flex.model.model import FlexModel
from flex.pool.pool import FlexPool


@pytest.fixture(name="fld")
def fixture_flex_dataset():
    """Function that returns a FlexDataset provided as example to test functions.

    Returns:
        FedDataset: A FlexDataset generated randomly
    """
    X_data = np.random.rand(100).reshape([20, 5])
    y_data = np.random.choice(2, 20)
    fcd = Dataset.from_array(X_data, y_data)
    X_data = np.random.rand(100).reshape([20, 5])
    y_data = np.random.choice(2, 20)
    fcd1 = Dataset.from_array(X_data, y_data)
    X_data = np.random.rand(100).reshape([20, 5])
    y_data = np.random.choice(2, 20)
    fcd2 = Dataset.from_array(X_data, y_data)
    return FedDataset({"client_1": fcd, "client_2": fcd1, "client_3": fcd2})


@pytest.fixture(name="only_clients")
def fixture_only_clients():
    client1_id = "client_1"
    role1 = FlexRole.client
    client2_id = "client_2"
    role2 = FlexRole.client
    client3_id = "client_3"
    role3 = FlexRole.client
    return FlexActors(
        {
            client1_id: role1,
            client2_id: role2,
            client3_id: role3,
        }
    )


class TestFlexPool(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def _fixture_only_clients(self, only_clients):
        self._only_clients = only_clients

    @pytest.fixture(autouse=True)
    def _fixture_iris_dataset(self):
        iris = load_iris()
        tmp = Dataset.from_array(iris.data, iris.target)
        self._iris = FedDataDistribution.iid_distribution(tmp, n_nodes=2)

    @pytest.fixture(autouse=True)
    def _fixture_flex_dataset(self, fld):
        self._fld = fld

    def test_len_property(self):
        p = FlexPool.client_server_pool(self._iris, lambda *args: None)
        assert len(p) != len(self._iris)
        assert len(p.select(lambda *args: True)) == len(p)
        assert len(p.actor_ids) == len(p)

    def test_iter_property(self):
        p = FlexPool.client_server_pool(self._iris, lambda *args: None)
        # Do it twice to ensure that the iterator is resetted properly
        assert all(a == b for a, b in zip(p._actors, p))
        assert len(p) == len(list(p))
        assert all(a == b for a, b in zip(p._actors, p))
        assert len(p) == len(list(p))

    def test_check_compatibility(self):
        p = FlexPool.client_server_pool(self._fld, lambda *args: None)
        server_pool = p.servers
        client_pool = p.clients
        assert FlexPool.check_compatibility(server_pool, client_pool) is True
        assert (
            FlexPool.check_compatibility(client_pool, server_pool) is True
        )  # Servers are also aggregators in this architecture
        assert FlexPool.check_compatibility(client_pool, client_pool) is False

    def test_map(self):
        p = FlexPool.client_server_pool(self._fld, lambda *args: None)
        server_pool = p.servers
        client_pool = p.clients
        assert server_pool.map(lambda *args: True, client_pool) == [True]
        assert client_pool.map(lambda *args: True) == len(client_pool) * [True]
        with pytest.raises(ValueError):
            assert client_pool.map(lambda *args: True, client_pool)

    def test_client_server_pool(self):
        p = FlexPool.client_server_pool(self._fld, lambda *args: None)
        assert FlexRoleManager.is_client(p._actors["client_1"]) is True
        assert FlexRoleManager.is_aggregator(p._actors["client_1"]) is False
        assert FlexRoleManager.is_server(p._actors["client_1"]) is False

        assert FlexRoleManager.is_client(p._actors["client_2"]) is True
        assert FlexRoleManager.is_aggregator(p._actors["client_2"]) is False
        assert FlexRoleManager.is_server(p._actors["client_2"]) is False

        assert FlexRoleManager.is_client(p._actors["client_3"]) is True
        assert FlexRoleManager.is_aggregator(p._actors["client_3"]) is False
        assert FlexRoleManager.is_server(p._actors["client_3"]) is False

        assert FlexRoleManager.is_client(p._actors["server"]) is False
        assert FlexRoleManager.is_aggregator(p._actors["server"]) is True
        assert FlexRoleManager.is_server(p._actors["server"]) is True

    def test_p2p_pool(self):
        p = FlexPool.p2p_pool(self._fld, lambda *args: None)
        assert FlexRoleManager.is_client(p._actors["client_1"]) is True
        assert FlexRoleManager.is_aggregator(p._actors["client_1"]) is True
        assert FlexRoleManager.is_server(p._actors["client_1"]) is True

        assert FlexRoleManager.is_client(p._actors["client_2"]) is True
        assert FlexRoleManager.is_aggregator(p._actors["client_2"]) is True
        assert FlexRoleManager.is_server(p._actors["client_2"]) is True

        assert FlexRoleManager.is_client(p._actors["client_3"]) is True
        assert FlexRoleManager.is_aggregator(p._actors["client_3"]) is True
        assert FlexRoleManager.is_server(p._actors["client_3"]) is True

    def test_validate_client_without_data(self):
        fld = FedDataset(
            {"client_1": self._fld["client_1"], "client_2": self._fld["client_2"]}
        )
        with pytest.raises(ValueError):
            FlexPool(fld, self._only_clients, None)

    def test_validate_client_without_all_models(self):
        models = defaultdict(None, {"client_1": 0, "client_2": 0})
        with pytest.raises(ValueError):
            FlexPool(self._fld, self._only_clients, models)

    def test_validate_client_without_role(self):
        client1_id = "client_1"
        role1 = FlexRole.client
        client2_id = "client_2"
        role2 = FlexRole.client
        missing_one = FlexActors(
            {
                client1_id: role1,
                client2_id: role2,
            }
        )
        with pytest.raises(ValueError):
            FlexPool(self._fld, missing_one, defaultdict())

    def test_validate_client_without_model(self):
        p = FlexPool.p2p_pool(self._fld, init_func=lambda *args: None)
        assert p._models.get("client_1") == FlexModel()
        assert p._models.get("client_2") == FlexModel()
        assert p._models.get("client_3") == FlexModel()

    def test_validate_flex_model_ids_in_pool_init(self):
        client1_id = "client_1"
        role1 = FlexRole.client
        client2_id = "client_2"
        role2 = FlexRole.client
        client3_id = "client_3"
        role3 = FlexRole.client
        actors = FlexActors(
            {
                client1_id: role1,
                client2_id: role2,
                client3_id: role3,
            }
        )
        flex_models = {k: FlexModel() for k in actors}
        for k in flex_models:
            flex_models[k].actor_id = 8
        with pytest.raises(ValueError):
            FlexPool(self._fld, actors, flex_models)

    def test_select(self):
        p = FlexPool.p2p_pool(self._fld, init_func=lambda *args: None)
        new_p = p.select(lambda a, b: FlexRoleManager.is_client(b))
        assert all(
            FlexRoleManager.is_client(actor_role)
            for _, actor_role in new_p._actors.items()
        )

    def test_select_int_criteria(self):
        iris = load_iris()
        tmp = Dataset.from_array(iris.data, iris.target)
        self._iris_many_clients = FedDataDistribution.iid_distribution(tmp, n_nodes=100)
        p = FlexPool.client_server_pool(self._iris, lambda *args: None)
        pool_size = len(p)
        assert all(len(p.select(i)) == i for i in range(pool_size))

    def test_client_property(self):
        p = FlexPool.p2p_pool(self._fld, init_func=lambda *args: None)
        new_p = p.clients
        assert all(
            FlexRoleManager.is_client(actor_role)
            for _, actor_role in new_p._actors.items()
        )

    def test_aggregator_property(self):
        p = FlexPool.p2p_pool(self._fld, init_func=lambda *args: None)
        new_p = p.aggregators
        assert all(
            FlexRoleManager.is_aggregator(actor_role)
            for _, actor_role in new_p._actors.items()
        )

    def test_server_property(self):
        p = FlexPool.p2p_pool(self._fld, init_func=lambda *args: None)
        new_p = p.servers
        assert all(
            FlexRoleManager.is_server(actor_role)
            for _, actor_role in new_p._actors.items()
        )



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Callable, List

import tensorly as tl
from numpy import ndarray

from flex.data.dataset import Dataset
from flex.distributed.client import Client
from flex.model import FlexModel
from flex.pool.aggregators import set_tensorly_backend


class FlexibleClient(Client):
    def __init__(
        self,
        flex_model: FlexModel,
        dataset: Dataset,
        train: Callable,
        collect: Callable,
        set_weights: Callable,
        eval: Callable,
        eval_dataset: Dataset,
    ) -> None:
        self._train = train
        self._collect = collect
        self._setweights = set_weights
        self._eval = eval
        super().__init__(dataset=dataset, model=flex_model, eval_dataset=eval_dataset)

    def train(self, model: FlexModel, data: Dataset):
        return self._train(model, data)

    def eval(self, model: FlexModel, data: Dataset):
        return self._eval(model, data)

    def get_weights(self, model: FlexModel) -> List[ndarray]:
        return self._collect(model)

    def set_weights(self, model: FlexModel, weights: List[ndarray]):
        return self._setweights(model, weights)


class ClientBuilder:
    """
    A builder class for creating a client that allows reusing the wrapped FLEX functions.

    This class provides a convenient way to build a client object by setting the necessary components
    such as the flex model, dataset, training function, evaluation function, and other required functions.

    Example usage:
    ```
    client = ClientBuilder()
        .model(flex_model)
        .dataset(dataset)
        .train(train_function)
        .collect_weights(collect_weights_function)
        .set_weights(set_weights_function)
        .eval(eval_function, eval_dataset)
        .build()
    ```

    Once all the required components are set, the `build()` method can be called to create a `Client` object.

    Note: It is important to set all the required components before calling the `build()` method.

    """

    def __init__(self) -> None:
        self._flex_model = None
        self._dataset = None
        self._train = None
        self._collect = None
        self._setweights = None
        self._eval = None
        self._eval_dataset = None

    def train(self, train: Callable) -> "ClientBuilder":
        """
        Set the training function for the client.

        Args:
        ----
            train (Callable): The training function.

        Returns:
        -------
            ClientBuilder: The updated client builder object.

        """
        self._train = train
        return self

    def eval(self, eval: Callable, eval_dataset: Dataset) -> "ClientBuilder":
        """
        Set the evaluation function and evaluation dataset for the client.

        Args:
        ----
            eval (Callable): The evaluation function.
            eval_dataset (Dataset): The evaluation dataset.

        Returns:
        -------
            ClientBuilder: The updated client builder object.

        """
        self._eval = eval
        self._eval_dataset = eval_dataset
        return self

    def collect_weights(self, collect_weights: Callable) -> "ClientBuilder":
        """
        Set the collect weights function for the client.

        Args:
        ----
            collect_weights (Callable): The collect weights function.

        Returns:
        -------
            ClientBuilder: The updated client builder object.

        Raises:
        ------
            ValueError: If the function passed as an argument is not wrapped by `collect_clients_weights`.

        """
        try:
            self._collect = collect_weights.__wrapped__

            def _collect_to_numpy_(model: FlexModel):
                weights = collect_weights.__wrapped__(model)
                set_tensorly_backend(weights)
                return [tl.to_numpy(tl.tensor(t)) for t in weights]

            self._collect = _collect_to_numpy_
            return self

        except AttributeError:
            raise ValueError(
                "Function passed as argument must be wrapped by collect_clients_weights"
            ) from None

    def set_weights(self, set_weights: Callable) -> "ClientBuilder":
        """
        Set the set weights function for the client. The recieved weights are numpy `ndarray`.

        Args:
        ----
            set_weights (Callable): The set weights function.

        Returns:
        -------
            ClientBuilder: The updated client builder object.

        Raises:
        ------
            ValueError: If the function passed as an argument is not wrapped by `set_aggregated_weights`.

        """
        try:
            self._setweights = set_weights.__wrapped__
            return self
        except AttributeError:
            raise ValueError(
                "Function passed as argument must be wrapped by set_aggregated_weights"
            ) from None

    def dataset(self, dataset: Dataset) -> "ClientBuilder":
        """
        Set the dataset for the client.

        Args:
        ----
            dataset (Dataset): The dataset.

        Returns:
        -------
            ClientBuilder: The updated client builder object.

        """
        self._dataset = dataset
        return self

    def model(self, flex_model: FlexModel) -> "ClientBuilder":
        """
        Set the flex model for the client.

        Args:
        ----
            flex_model (FlexModel): The flex model.

        Returns:
        -------
            ClientBuilder: The updated client builder object.

        """
        self._flex_model = flex_model
        return self

    def build_model(self, build_server_model: Callable) -> "ClientBuilder":
        """
        Builds the client model by calling the provided `build_server_model` function.

        Args:
        ----
            build_server_model (Callable): A function that builds the server model.

        Returns:
        -------
            ClientBuilder: The instance of the ClientBuilder class.

        """
        return self.model(build_server_model.__wrapped__())

    def build(self) -> Client:
        """
        Build and return the client object.

        Returns
        -------
            Client: The client object.

        Raises
        ------
            AssertionError: If any of the required components are not set.

        """
        assert self._flex_model is not None, "model must be set"
        assert self._dataset is not None, "dataset must be set"
        assert self._train is not None, "train function must be set"
        assert self._collect is not None, "collect function must be set"
        assert self._setweights is not None, "set_weights function must be set"
        assert self._eval is not None, "eval function must be set"
        assert self._eval_dataset is not None, "eval dataset must be set"
        return FlexibleClient(
            self._flex_model,
            self._dataset,
            self._train,
            self._collect,
            self._setweights,
            self._eval,
            self._eval_dataset,
        )



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import copy
import unittest

import pytest
from sklearn.datasets import load_iris
from sklearn.metrics import accuracy_score
from sklearn.neighbors import KNeighborsClassifier

from flex.data import FedDataDistribution
from flex.data.fed_dataset import Dataset
from flex.model.model import FlexModel
from flex.pool.decorators import (
    aggregate_weights,
    collect_clients_weights,
    deploy_server_model,
    evaluate_server_model,
    init_server_model,
    set_aggregated_weights,
)
from flex.pool.pool import FlexPool


class TestFlexPool(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def _fixture_iris_dataset(self):
        iris = load_iris()
        c_iris = Dataset.from_array(iris.data, iris.target)
        self.f_iris = FedDataDistribution.iid_distribution(c_iris, n_nodes=2)

    def test_decorators_guard(self):
        with pytest.raises(AssertionError):

            @set_aggregated_weights
            def bad_set_agreggated_weights_to_server(server_flex_model):
                pass

    def test_decorators(self):
        @init_server_model
        def build_server_model():
            flex_model = FlexModel()
            flex_model["model"] = KNeighborsClassifier(n_neighbors=3)
            return flex_model

        @deploy_server_model
        def copy_server_model_to_clients(server_flex_model):
            flex_model = FlexModel()
            for k, v in server_flex_model.items():
                flex_model[k] = copy.deepcopy(v)

            return flex_model

        @collect_clients_weights
        def get_clients_weights(client_flex_model):
            return client_flex_model["model"].get_params()

        @aggregate_weights
        def aggregate(list_of_weights):
            return sum(w["leaf_size"] for w in list_of_weights)

        @set_aggregated_weights
        def set_agreggated_weights_to_server(server_flex_model, aggregated_weights):
            server_flex_model["model"].set_params(leaf_size=aggregated_weights)

        @evaluate_server_model
        def evaluate_server(server_flex_model, test_data=None):
            x, y = test_data.to_numpy()
            server_flex_model["model"].fit(x, y)
            preds = server_flex_model["model"].predict(x)
            return preds, accuracy_score(preds, y)

        p = FlexPool.client_server_pool(self.f_iris, init_func=build_server_model)
        reference_model_params = KNeighborsClassifier(n_neighbors=3).get_params()
        reference_value = reference_model_params["leaf_size"] * len(p.clients)

        p.servers.map(copy_server_model_to_clients, p.clients)
        assert all(
            p._models[k]["model"].get_params() == reference_model_params
            for k in p.actor_ids
        )
        p.aggregators.map(get_clients_weights, p.clients)
        # Aggregate weights
        p.aggregators.map(aggregate)
        assert p._models["server"]["aggregated_weights"] == reference_value
        # Transfer weights from aggregators to servers
        p.aggregators.map(set_agreggated_weights_to_server, p.servers)
        # Deploy new model to clients
        p.servers.map(copy_server_model_to_clients, p.clients)
        assert all(
            p._models[k]["model"].get_params()["leaf_size"] == reference_value
            for k in p.actor_ids
        )
        test_data = self.f_iris[0]
        result = p.servers.map(evaluate_server, test_data=test_data)
        preds, accuracy = result[0]
        print(accuracy)
        assert len(preds) == len(test_data)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: transport.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import tensor_pb2 as tensor__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0ftransport.proto\x12\x08\x66lexible\x1a\x0ctensor.proto"\x17\n\x05\x45rror\x12\x0e\n\x06reason\x18\x01 \x01(\t"\xad\x06\n\rClientMessage\x12=\n\rhandshake_res\x18\x01 \x01(\x0b\x32$.flexible.ClientMessage.HandshakeResH\x00\x12@\n\x0fget_weights_res\x18\x02 \x01(\x0b\x32%.flexible.ClientMessage.GetWeightsResH\x00\x12\x42\n\x10send_weights_res\x18\x03 \x01(\x0b\x32&.flexible.ClientMessage.SendWeightsResH\x00\x12\x35\n\ttrain_res\x18\x04 \x01(\x0b\x32 .flexible.ClientMessage.TrainResH\x00\x12 \n\x05\x65rror\x18\x05 \x01(\x0b\x32\x0f.flexible.ErrorH\x00\x12\x33\n\x08\x65val_res\x18\x06 \x01(\x0b\x32\x1f.flexible.ClientMessage.EvalResH\x00\x12\x38\n\nhealth_ins\x18\x07 \x01(\x0b\x32".flexible.ClientMessage.HealthPingH\x00\x1a\x1e\n\x0cHandshakeRes\x12\x0e\n\x06status\x18\x01 \x01(\x05\x1a\x32\n\rGetWeightsRes\x12!\n\x07weights\x18\x01 \x03(\x0b\x32\x10.flexible.Tensor\x1a \n\x0eSendWeightsRes\x12\x0e\n\x06status\x18\x01 \x01(\x05\x1az\n\x08TrainRes\x12>\n\x07metrics\x18\x01 \x03(\x0b\x32-.flexible.ClientMessage.TrainRes.MetricsEntry\x1a.\n\x0cMetricsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\x1ax\n\x07\x45valRes\x12=\n\x07metrics\x18\x01 \x03(\x0b\x32,.flexible.ClientMessage.EvalRes.MetricsEntry\x1a.\n\x0cMetricsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\x1a\x1c\n\nHealthPing\x12\x0e\n\x06status\x18\x01 \x01(\x05\x42\x05\n\x03msg"\x8f\x04\n\rServerMessage\x12@\n\x0fget_weights_ins\x18\x01 \x01(\x0b\x32%.flexible.ServerMessage.GetWeightsInsH\x00\x12\x42\n\x10send_weights_ins\x18\x02 \x01(\x0b\x32&.flexible.ServerMessage.SendWeightsInsH\x00\x12\x35\n\ttrain_ins\x18\x03 \x01(\x0b\x32 .flexible.ServerMessage.TrainInsH\x00\x12 \n\x05\x65rror\x18\x04 \x01(\x0b\x32\x0f.flexible.ErrorH\x00\x12\x33\n\x08\x65val_ins\x18\x05 \x01(\x0b\x32\x1f.flexible.ServerMessage.EvalInsH\x00\x12\x38\n\nhealth_ins\x18\x06 \x01(\x0b\x32".flexible.ServerMessage.HealthPingH\x00\x1a\x1f\n\rGetWeightsIns\x12\x0e\n\x06status\x18\x01 \x01(\x05\x1a\x33\n\x0eSendWeightsIns\x12!\n\x07weights\x18\x01 \x03(\x0b\x32\x10.flexible.Tensor\x1a\x1a\n\x08TrainIns\x12\x0e\n\x06status\x18\x01 \x01(\x05\x1a\x19\n\x07\x45valIns\x12\x0e\n\x06status\x18\x01 \x01(\x05\x1a\x1c\n\nHealthPing\x12\x0e\n\x06status\x18\x01 \x01(\x05\x42\x05\n\x03msg2H\n\x08\x46lexible\x12<\n\x04Send\x12\x17.flexible.ClientMessage\x1a\x17.flexible.ServerMessage(\x01\x30\x01\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "transport_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_CLIENTMESSAGE_TRAINRES_METRICSENTRY"]._loaded_options = None
    _globals["_CLIENTMESSAGE_TRAINRES_METRICSENTRY"]._serialized_options = b"8\001"
    _globals["_CLIENTMESSAGE_EVALRES_METRICSENTRY"]._loaded_options = None
    _globals["_CLIENTMESSAGE_EVALRES_METRICSENTRY"]._serialized_options = b"8\001"
    _globals["_ERROR"]._serialized_start = 43
    _globals["_ERROR"]._serialized_end = 66
    _globals["_CLIENTMESSAGE"]._serialized_start = 69
    _globals["_CLIENTMESSAGE"]._serialized_end = 882
    _globals["_CLIENTMESSAGE_HANDSHAKERES"]._serialized_start = 483
    _globals["_CLIENTMESSAGE_HANDSHAKERES"]._serialized_end = 513
    _globals["_CLIENTMESSAGE_GETWEIGHTSRES"]._serialized_start = 515
    _globals["_CLIENTMESSAGE_GETWEIGHTSRES"]._serialized_end = 565
    _globals["_CLIENTMESSAGE_SENDWEIGHTSRES"]._serialized_start = 567
    _globals["_CLIENTMESSAGE_SENDWEIGHTSRES"]._serialized_end = 599
    _globals["_CLIENTMESSAGE_TRAINRES"]._serialized_start = 601
    _globals["_CLIENTMESSAGE_TRAINRES"]._serialized_end = 723
    _globals["_CLIENTMESSAGE_TRAINRES_METRICSENTRY"]._serialized_start = 677
    _globals["_CLIENTMESSAGE_TRAINRES_METRICSENTRY"]._serialized_end = 723
    _globals["_CLIENTMESSAGE_EVALRES"]._serialized_start = 725
    _globals["_CLIENTMESSAGE_EVALRES"]._serialized_end = 845
    _globals["_CLIENTMESSAGE_EVALRES_METRICSENTRY"]._serialized_start = 677
    _globals["_CLIENTMESSAGE_EVALRES_METRICSENTRY"]._serialized_end = 723
    _globals["_CLIENTMESSAGE_HEALTHPING"]._serialized_start = 847
    _globals["_CLIENTMESSAGE_HEALTHPING"]._serialized_end = 875
    _globals["_SERVERMESSAGE"]._serialized_start = 885
    _globals["_SERVERMESSAGE"]._serialized_end = 1412
    _globals["_SERVERMESSAGE_GETWEIGHTSINS"]._serialized_start = 1236
    _globals["_SERVERMESSAGE_GETWEIGHTSINS"]._serialized_end = 1267
    _globals["_SERVERMESSAGE_SENDWEIGHTSINS"]._serialized_start = 1269
    _globals["_SERVERMESSAGE_SENDWEIGHTSINS"]._serialized_end = 1320
    _globals["_SERVERMESSAGE_TRAININS"]._serialized_start = 1322
    _globals["_SERVERMESSAGE_TRAININS"]._serialized_end = 1348
    _globals["_SERVERMESSAGE_EVALINS"]._serialized_start = 1350
    _globals["_SERVERMESSAGE_EVALINS"]._serialized_end = 1375
    _globals["_SERVERMESSAGE_HEALTHPING"]._serialized_start = 847
    _globals["_SERVERMESSAGE_HEALTHPING"]._serialized_end = 875
    _globals["_FLEXIBLE"]._serialized_start = 1414
    _globals["_FLEXIBLE"]._serialized_end = 1486
# @@protoc_insertion_point(module_scope)



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import copy
from collections import defaultdict
from math import floor
from typing import Callable, Tuple

import numpy as np
import numpy.typing as npt
from numpy.random import default_rng

from flex.data import Dataset, FedDataset, FedDatasetConfig
from flex.data.lazy_indexable import LazyIndexable


class FedDataDistribution(object):
    __create_key = object()

    def __init__(self, create_key: object = None) -> None:
        assert (
            create_key == FedDataDistribution.__create_key
        ), """FedDataDistribution objects must be created using FedDataDistribution.from_config or
        FedDataDistribution.iid_distribution"""

    @classmethod
    def from_config_with_torchtext_dataset(cls, data, config: FedDatasetConfig):
        """This function federates a centralized torchtext dataset given a FlexDatasetConfig.
        This function will transform the torchtext dataset into a Dataset and then it will
        federate it.

        Args:
        -----
            data (Dataset): The torchtext dataset
            config (FedDatasetConfig): FlexDatasetConfig with the configuration to federate the centralized dataset.
        """
        centralized_data = Dataset.from_torchtext_dataset(data)
        return cls.from_config(centralized_data, config)

    @classmethod
    def from_config_with_tfds_image_dataset(cls, data, config: FedDatasetConfig):
        """This function federates a centralized tensorflow dataset given a FlexDatasetConfig.
        This function will transform a dataset from the tensorflow_datasets module into a Dataset
        and then it will federate it.

        Args:
        -----
            data (Dataset): The tensorflow dataset
            config (FedDatasetConfig): FlexDatasetConfig with the configuration to federate the centralized dataset.
        """
        centralized_data = Dataset.from_tfds_image_dataset(data)
        return cls.from_config(centralized_data, config)

    @classmethod
    def from_config_with_tfds_text_dataset(
        cls, data, config: FedDatasetConfig, X_columns: list, label_columns: list
    ):
        """This function federates a centralized tensorflow dataset given a FlexDatasetConfig.
        This function will transform a dataset from the tensorflow_datasets module into a Dataset
        and then it will federate it.

        Args:
        -----
            data (Dataset): The tensorflow dataset
            config (FedDatasetConfig): FlexDatasetConfig with the configuration to federate the centralized dataset.
            X_columns (List): List that contains the columns names for the input features.
            label_columns (List): List that contains the columns names for the output features.
        """
        centralized_data = Dataset.from_tfds_text_dataset(
            data, X_columns, label_columns
        )
        return cls.from_config(centralized_data, config)

    @classmethod
    def from_config_with_torchvision_dataset(cls, data, config: FedDatasetConfig):
        """This function federates a centralized torchvision dataset given a FlexDatasetConfig.
        This function will transform a dataset from the torchvision module into a Dataset
        and then it will federate it.

        Args:
        -----
            data (Dataset): The torchvision dataset
            config (FedDatasetConfig): FlexDatasetConfig with the configuration to federate the centralized dataset.
        """
        centralized_data = Dataset.from_torchvision_dataset(data)
        return cls.from_config(centralized_data, config)

    @classmethod
    def from_config_with_huggingface_dataset(
        cls,
        data,
        config: FedDatasetConfig,
        X_columns: list,
        label_columns: list = None,
    ):
        """This function federates a centralized hugginface dataset given a FlexDatasetConfig.
        This function will transform a dataset from the HuggingFace Hub datasets into a Dataset
        and then it will federate it.

        Args:
        -----
            data (Union[datasets.arrow_dataset.Dataset, str]): The hugginface dataset to federate.
            config (FedDatasetConfig): FlexDatasetConfig with the configuration to federate the centralized dataset.
            X_coluns (List[str]): List with the names of the columns to load.
            label_columns (list): List with the names of the label columns.
        """
        centralized_data = Dataset.from_huggingface_dataset(
            data, X_columns, label_columns
        )
        return cls.from_config(centralized_data, config)

    @classmethod
    def from_clustering_func(cls, centralized_data: Dataset, clustering_func: Callable):
        """This function federates data into nodes by means of a clustering function, that outputs
        to which node (cluster) a data point belongs.

        Args:
        -----
            centralized_data (Dataset): Centralized dataset represented as a FlexDataObject.
            clustering_func (Callable): function that receives as arguments a pair of x and y elements from centralized_data
            and returns the name of the node (cluster) that should own it, the returned type must be Hashable.
            Note that we only support one node (cluster) per data point.

        Returns:
        --------
            federated_dataset (FedDataset): The federated dataset.
        """
        d = defaultdict(list)
        for idx, (x, y) in enumerate(centralized_data):
            node_name = clustering_func(x, y)
            d[node_name].append(idx)

        config = FedDatasetConfig(
            n_nodes=len(d),
            node_ids=list(d.keys()),
            indexes_per_node=list(d.values()),
            replacement=False,
        )
        return cls.from_config(centralized_data, config)

    @classmethod
    def iid_distribution(cls, centralized_data: Dataset, n_nodes: int = 2):
        """Function to create a FedDataset for an IID experiment. We consider the simplest situation
        in which the data is distributed by giving the same amount of data to each node.

        Args:
        -----
            centralized_data (Dataset): Centralized dataset represented as a FlexDataObject.
            n_nodes (int): Number of nodes in the Federated Learning experiment. Default 2.

        Returns:
        --------
            federated_dataset (FedDataset): The federated dataset.
        """
        config = FedDatasetConfig(n_nodes=n_nodes)
        return FedDataDistribution.from_config(centralized_data, config)

    @classmethod
    def from_config(cls, centralized_data: Dataset, config: FedDatasetConfig):
        """This function prepare the data from a centralized data structure to a federated one.
        It will run different modifications to federate the data.

        Args:
        -----
            centralized_data (Dataset): Centralized dataset represented as a FlexDataObject.
            config (FedDatasetConfig): FlexDatasetConfig with the configuration to federate the centralized dataset.

        Returns:
        --------
            federated_dataset (FedDataset): The federated dataset.
        """
        centralized_data.validate()
        config.validate()
        rng = default_rng(seed=config.seed)

        config_ = copy.deepcopy(config)  # copy, because we might modify some components

        if config.node_ids is None:
            config_.node_ids = list(range(config_.n_nodes))

        if config.keep_labels is None:
            config_.keep_labels = [True] * config_.n_nodes

        labels = None
        if centralized_data.y_data is not None:
            labels = centralized_data.y_data.to_numpy()

        # Normalize weights when no replacement
        if (
            not config_.replacement
            and config_.weights is not None
            and sum(config_.weights) > 1
        ):
            config_.weights = np.array(
                [w / sum(config.weights) for w in config.weights]
            )
        # Ensure that labels_per_node is translated to weights_per_label
        if config_.labels_per_node is not None:
            cls.__configure_weights_per_class(rng, config_, labels)
        # Normalize weights_per_label when no replacement
        if (
            not config_.replacement
            and config_.weights_per_label is not None
            and any(np.sum(config_.weights_per_label, axis=0) > 1)
        ):
            with np.errstate(divide="ignore", invalid="ignore"):  # raise no warnings
                config_.weights_per_label = config_.weights_per_label / np.sum(
                    config_.weights_per_label, axis=0
                )
            # Note that weights equal to 0 produce NaNs, so we replace them with 0 again
            config_.weights_per_label = np.nan_to_num(config_.weights_per_label)

        # Now we can start generating our federated dataset
        fed_dataset = FedDataset()
        if config_.indexes_per_node is not None:
            for node_name, data in cls.__sample_dataset_with_indexes(
                centralized_data, config_
            ):
                fed_dataset[node_name] = data
        elif config_.group_by_label_index is not None:
            for node_name, data in cls.__group_by_label_index(
                centralized_data, config_
            ):
                fed_dataset[node_name] = data
        else:  # sample using weights or features
            remaining_data_indices = np.arange(len(labels))
            for i in range(config_.n_nodes):
                if config_.shuffle:
                    rng.shuffle(remaining_data_indices)
                keep_y_data = (
                    centralized_data.y_data is not None and config_.keep_labels[i]
                )
                (
                    sub_data_indices,
                    sub_features_indices,
                    remaining_data_indices,
                ) = cls.__sample(
                    rng, remaining_data_indices, centralized_data, labels, config_, i
                )
                X_data = centralized_data.X_data[sub_data_indices]
                if config.features_per_node is not None:
                    X_data = X_data.to_numpy()
                    X_data = LazyIndexable(X_data[:, sub_features_indices], len(X_data))
                if keep_y_data:
                    y_data = centralized_data.y_data[sub_data_indices]
                else:
                    y_data = None
                fed_dataset[config_.node_ids[i]] = Dataset(X_data=X_data, y_data=y_data)

        return fed_dataset

    @classmethod
    def __group_by_label_index(
        cls, centralized_data: Dataset, config: FedDatasetConfig
    ):
        label_index = config.group_by_label_index
        label_to_node_id = {}
        y_data = defaultdict(list)
        x_data_indexes = defaultdict(list)
        for i, y in enumerate(centralized_data.y_data):
            y = list(y)  # TODO: enforce that y is only a list or a tuple
            str_label = str(y.pop(label_index))  # Use str to make every label hashable
            if str_label not in label_to_node_id:
                label_to_node_id[
                    str_label
                ] = i  # Name each node using the first index where the label appears
            x_data_indexes[label_to_node_id[str_label]].append(i)
            if len(y) == 1:
                y = y[0]
            y_data[label_to_node_id[str_label]].append(y)
        for node_id in y_data:
            yield node_id, Dataset(
                X_data=centralized_data.X_data[x_data_indexes[node_id]],
                y_data=LazyIndexable(y_data[node_id], len(y_data[node_id])),
            )

    @classmethod
    def __sample_dataset_with_indexes(cls, data: Dataset, config: FedDatasetConfig):
        """Iterable function that associates a node with its data, when a list of indexes is given for
        each node.

        Args:
        -----
            data (Dataset): Centralizaed dataset represented as a FlexDataObject.
            config (FedDatasetConfig): Configuration used to federate a FlexDataObject.

        Yields:
        -------
            tuple (Tuple): a tuple whose first item is the node name and the second one is the indexes of
            the dataset associated to such node.

        """
        for idx, name, keep in zip(
            config.indexes_per_node, config.node_ids, config.keep_labels
        ):
            yield name, Dataset(
                X_data=data.X_data[idx],
                y_data=data.y_data[idx] if data.y_data is not None and keep else None,
            )

    @classmethod
    def __sample(
        cls,
        rng: np.random.Generator,
        data_indices: npt.NDArray[np.int_],
        data: Dataset,
        labels: npt.ArrayLike,
        config: FedDatasetConfig,
        node_i: int,
    ) -> Tuple[npt.NDArray[np.int_], npt.NDArray[np.int_], npt.NDArray[np.int_]]:
        """Function to sample indices from a FlexDataObject as especified by a FlexDatasetConfig.

        Args:
        -----
            rng (np.random.Generator): Random number generator used to sample.
            data_indices (npt.NDArray[np.int_]): Array of available data indices to sample from.
            data (Dataset): Centralizaed dataset represented as a FlexDataObject.
            config (FedDatasetConfig): Configuration used to federate a FlexDataObject.
            node_i (int): Position of node which will be identified with the generated sample.

        Returns:
        --------
            sample_indices (Tuple[npt.NDArray[np.int_], npt.NDArray[np.int_], npt.NDArray[np.int_]]): it returns
            the sampled data indices, the sampled feature indices and the data indices which were not used for
            the sampled data indices. Note that, the latter are only used for the config.replacement option, otherwise
            it contains all the provided data_indices.
        """
        # Sample feature indices
        sub_features_indices = cls.__sample_features(rng, data, config, node_i)
        # Sample data indices
        sub_data_indices = cls.__sample_with_weights(
            rng, data_indices, labels, config, node_i
        )
        # Update remaining data indices
        remaining_data_indices = (
            data_indices
            if config.replacement
            else np.array(list(set(data_indices) - set(sub_data_indices)))
        )

        return sub_data_indices, sub_features_indices, remaining_data_indices

    @classmethod
    def __sample_with_weights(
        cls,
        rng: np.random.Generator,
        data_indices: npt.NDArray[np.int_],
        labels: npt.ArrayLike,
        config: FedDatasetConfig,
        node_i: int,
    ):
        """Especialized function to sample indices from a FlexDataObject as especified by a FlexDatasetConfig.
            It takes into consideration the config.weights and config.weights_per_label option and applies it.
            If no config.weights and no config.weights_per_label are provided, then we consider that the weights \
            are the same for all the nodes.

        Args:
        -----
            rng (np.random.Generator): Random number generator used to sample.
            data_indices (npt.NDArray[np.int_]): Array of available data indices to sample from.
            data (Dataset): Centralizaed dataset represented as a FlexDataObject.
            config (FedDatasetConfig): Configuration used to federate a FlexDataObject.
            node_i (int): Position of node which will be identified with the generated sample.

        Returns:
        --------
            sample_indices (Tuple[npt.NDArray[np.int_], npt.NDArray[np.int_]]): it returns
            the sampled data indices and all the feature indices.
        """
        if config.weights_per_label is not None:
            data_proportion = None
        elif config.weights is not None:
            data_proportion = floor(len(labels) * config.weights[node_i])
        else:  # No weights provided
            data_proportion = floor(len(labels) / config.n_nodes)

        if data_proportion is not None:
            sub_data_indices = data_indices[:data_proportion]
        else:  # apply weights_per_label
            sub_data_indices = np.array([], dtype="uint32")
            sorted_labels = np.sort(np.unique(labels))
            available_indices = copy.deepcopy(data_indices)
            if config.shuffle:
                rng.shuffle(available_indices)
            proportion_per_label = {}
            for j, label in enumerate(sorted_labels):
                available_class_indices = sum(labels == label)
                proportion_per_label[label] = floor(
                    available_class_indices * config.weights_per_label[node_i][j]
                )
            for label in proportion_per_label:
                available_class_indices = available_indices[
                    labels[available_indices] == label
                ]
                selected_class_indices = available_class_indices[
                    : proportion_per_label[label]
                ]
                sub_data_indices = np.concatenate(
                    (sub_data_indices, selected_class_indices)
                )
                available_indices = np.array(
                    list(set(available_indices) - set(selected_class_indices))
                )

        return sub_data_indices

    @classmethod
    def __configure_weights_per_class(
        cls, rng: np.random.Generator, config: FedDatasetConfig, labels: npt.ArrayLike
    ):
        sorted_labels = np.sort(np.unique(labels))
        assigned_labels = []
        if isinstance(config.labels_per_node, int):
            histogram = np.zeros_like(sorted_labels)
            for _ in range(config.n_nodes):
                individual_assigned_labels = []
                for _ in range(config.labels_per_node):
                    most_frequent = np.max(histogram)
                    available_labels_indexes = np.arange(len(sorted_labels))
                    tmp_available_indexes = histogram < most_frequent
                    if sum(tmp_available_indexes) != 0:
                        available_labels_indexes = available_labels_indexes[
                            tmp_available_indexes
                        ]
                    indx = rng.choice(available_labels_indexes, size=1, replace=False)
                    histogram[indx] = histogram[indx] + 1
                    individual_assigned_labels.append(sorted_labels[indx])
                assigned_labels.append(individual_assigned_labels)
            config.labels_per_node = assigned_labels
        elif isinstance(config.labels_per_node, tuple):
            num_labels_per_node = rng.integers(
                low=config.labels_per_node[0],
                high=config.labels_per_node[1] + 1,
                size=config.n_nodes,
            )
            for c in num_labels_per_node:
                n = rng.choice(sorted_labels, size=c, replace=False)
                assigned_labels.append(n)
            config.labels_per_node = assigned_labels

        config.weights_per_label = np.zeros((config.n_nodes, len(sorted_labels)))
        for node_i, clasess_at_node_i in enumerate(config.labels_per_node):
            for class_j, label in enumerate(sorted_labels):
                if label in clasess_at_node_i:
                    if config.weights is None:
                        config.weights_per_label[node_i, class_j] = 1
                    else:
                        config.weights_per_label[node_i, class_j] = config.weights[
                            node_i
                        ] / len(clasess_at_node_i)

    @classmethod
    def __sample_features(
        cls,
        rng: np.random.Generator,
        data: Dataset,
        config: FedDatasetConfig,
        node_i: int,
    ):
        """Especialized function to sample indices from a FlexDataObject as especified by a FlexDatasetConfig.
            It takes into consideration the config.features_per_node option and applies it.

        Args:
        -----
            rng (np.random.Generator): Random number generator used to sample.
            data (Dataset): Centralized dataset represented as a FlexDataObject.
            config (FedDatasetConfig): Configuration used to federate a FlexDataObject.
            node_i (int): Position of node which will be identified with the generated sample.

        Returns:
        --------
            sample_indices (Tuple[npt.NDArray[np.int_], npt.NDArray[np.int_]): it returns the sampled data indices
            and the sampled feature indices.
        """

        if config.features_per_node is None:
            sub_features_indices = slice(None)
        else:
            feature_indices = np.arange(len(data.X_data[0]))
            if isinstance(  # We have a fixed number of features per node
                config.features_per_node, int
            ):
                sub_features_indices = rng.choice(
                    feature_indices, config.features_per_node, replace=False
                )
            elif isinstance(  # We have a maximum and a minimum of features per node
                config.features_per_node, tuple
            ):
                sub_features_indices = rng.choice(
                    feature_indices,
                    rng.integers(
                        config.features_per_node[0], config.features_per_node[1] + 1
                    ),
                    replace=False,
                )
            else:  # We have an array of features per node, that is, each node has an set of labels
                sub_features_indices = config.features_per_node[node_i]

        return sub_features_indices



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np

from flex.data import Dataset, FedDataDistribution, FedDatasetConfig
from flex.datasets import standard_datasets


def federated_emnist(out_dir: str = ".", split="digits", return_test=False):
    train_data, _ = standard_datasets.emnist(out_dir, split=split, include_authors=True)
    config = FedDatasetConfig(
        group_by_label_index=1
    )  # when authoers are included, each label is a tuple (class, writer_id)
    federated_data = FedDataDistribution.from_config(train_data, config)
    if return_test:
        _, test_data = standard_datasets.emnist(
            out_dir, split=split, include_authors=False
        )
        return (federated_data, test_data)
    else:
        return federated_data


def federated_celeba(out_dir: str = ".", return_test=False):
    from torchvision.datasets import CelebA

    class ToNumpy:
        def __call__(self, data):
            if isinstance(data, tuple):  # Label
                return tuple(np.asarray(i) for i in data)
            else:
                return np.asarray(data)  # Images

    dataset = CelebA(
        root=out_dir,
        split="train",
        transform=ToNumpy(),
        target_transform=ToNumpy(),
        target_type=["identity", "attr"],
        download=True,
    )
    config = FedDatasetConfig(group_by_label_index=0)  # identity
    federated_data = FedDataDistribution.from_config_with_torchvision_dataset(
        dataset, config
    )
    if return_test:
        test_ds = CelebA(
            root=out_dir,
            split="test",
            transform=ToNumpy(),
            target_transform=ToNumpy(),
            target_type=["identity", "attr"],
            download=True,
        )
        test_data = Dataset.from_torchvision_dataset(test_ds)
        return (federated_data, test_data)
    return federated_data


def federated_sentiment140(out_dir: str = ".", return_test=False, **kwargs):
    from datasets import load_dataset

    dataset = load_dataset("sentiment140")
    x_labels = ["text"]
    y_labels = ["user", "sentiment"]
    config = FedDatasetConfig(group_by_label_index=0)  # Label "user"
    federated_data = FedDataDistribution.from_config_with_huggingface_dataset(
        dataset["train"], config, x_labels, y_labels
    )
    if return_test:
        test_data = Dataset.from_huggingface_dataset(
            dataset["test"], x_labels, y_labels
        )
        return (federated_data, test_data)
    return federated_data


def federated_shakespeare(out_dir: str = ".", return_test=False):
    train_data, _ = standard_datasets.shakespeare(out_dir, include_actors=True)
    config = FedDatasetConfig(
        group_by_label_index=1
    )  # each label is a pair (class, actor_id)
    federated_data = FedDataDistribution.from_config(train_data, config)
    if return_test:
        _, test_data = standard_datasets.shakespeare(out_dir, include_actors=False)
        return (federated_data, test_data)
    else:
        return federated_data



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
"""File that contains the adapted aggregators in FLEXible for fast
development of a federated model in FLEXible.

This aggregators also can work as examples for creating a custom aggregator.
"""

import tensorly as tl  # noqa: E402

from flex.pool.decorators import aggregate_weights  # noqa: E402


def flatten(xs):
    for x in xs:
        if isinstance(x, (list, tuple)):
            yield from flatten(x)
        else:
            yield x


def set_tensorly_backend(
    aggregated_weights_as_list: list, supported_modules: list = None
):  # jax support is planned
    if supported_modules is None:
        supported_modules = ["tensorflow", "torch"]
    backend_set = False
    for modulename in supported_modules:
        try:
            tmp_import = __import__(modulename)
            if all(
                tmp_import.is_tensor(t) for t in flatten(aggregated_weights_as_list)
            ):
                if modulename == "torch":
                    modulename = f"py{modulename}"
                tl.set_backend(modulename)
                backend_set = True
                break
            else:
                del tmp_import
        except ImportError:
            ...
    # Default backend
    if not backend_set:
        tl.set_backend("numpy")


def fed_avg_f(aggregated_weights_as_list: list):
    n_nodes = len(aggregated_weights_as_list)
    ponderation = [1 / n_nodes] * n_nodes
    return weighted_fed_avg_f(aggregated_weights_as_list, ponderation)


def weighted_fed_avg_f(aggregated_weights_as_list: list, ponderation: list):
    n_layers = len(aggregated_weights_as_list[0])
    agg_weights = []
    for layer_index in range(n_layers):
        weights_per_layer = []
        for client_weights, p in zip(aggregated_weights_as_list, ponderation):
            context = tl.context(client_weights[layer_index])
            w = client_weights[layer_index] * tl.tensor(p, **context)
            weights_per_layer.append(w)
        weights_per_layer = tl.stack(weights_per_layer)
        agg_layer = tl.sum(weights_per_layer, axis=0)
        agg_weights.append(agg_layer)
    return agg_weights


@aggregate_weights
def fed_avg(aggregated_weights_as_list: list):
    """Function that implements the FedAvg aggregation method

    Args:
    -----
        aggregated_weights_as_list (list): List which contains
        all the weights to aggregate

    Returns:
    --------
        tensor array: An array with the aggregated weights

    Example of use assuming you are using a client-server architecture:

        from flex.pool.primitive_functions import fed_avg

        aggregator = flex_pool.aggregators
        server = flex_pool.servers
        aggregator.map(server, fed_avg)

    Example of use using the FlexPool without separating server
    and aggregator, and following a client-server architecture.

        from flex.pool.primitive_functions import fed_avg

        flex_pool.aggregators.map(flex_pool.servers, fed_avg)
    """
    set_tensorly_backend(aggregated_weights_as_list)
    return fed_avg_f(aggregated_weights_as_list)


@aggregate_weights
def weighted_fed_avg(aggregated_weights_as_list: list, ponderation: list):
    """Function that implements the weighted FedAvg aggregation method.

    Args:
    -----
        aggregated_weights_as_list (list): List which contains
        all the weights to aggregate
        ponderation (list): weights assigned to each client

    Returns:
    --------
        tensor array: An array with the aggregated weights

    Example of use assuming you are using a client-server architecture:

        from flex.pool.primitive_functions import weighted_fed_avg

        aggregator = flex_pool.aggregators
        server = flex_pool.servers
        dummy_poderation = [1.]*len(flex_pool.clients)
        aggregator.map(server, weighted_fed_avg, ponderation=dummy_poderation)

    Example of use using the FlexPool without separating server
    and aggregator, and following a client-server architecture.

        from flex.pool.primitive_functions import weighted_fed_avg
        dummy_poderation = [1.]*len(flex_pool.clients)
        flex_pool.aggregators.map(flex_pool.servers, weighted_fed_avg, ponderation=dummy_poderation)
    """
    set_tensorly_backend(aggregated_weights_as_list)
    return weighted_fed_avg_f(aggregated_weights_as_list, ponderation)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import warnings

import grpc

from . import transport_pb2 as transport__pb2

GRPC_GENERATED_VERSION = "1.64.1"
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = "1.65.0"
SCHEDULED_RELEASE_DATE = "June 25, 2024"
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower

    _version_not_supported = first_version_is_lower(
        GRPC_VERSION, GRPC_GENERATED_VERSION
    )
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f"The grpc package installed is at version {GRPC_VERSION},"
        + f" but the generated code in transport_pb2_grpc.py depends on"
        + f" grpcio>={GRPC_GENERATED_VERSION}."
        + f" Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}"
        + f" or downgrade your generated code using grpcio-tools<={GRPC_VERSION}."
        + f" This warning will become an error in {EXPECTED_ERROR_RELEASE},"
        + f" scheduled for release on {SCHEDULED_RELEASE_DATE}.",
        RuntimeWarning,
    )


class FlexibleStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Send = channel.stream_stream(
            "/flexible.Flexible/Send",
            request_serializer=transport__pb2.ClientMessage.SerializeToString,
            response_deserializer=transport__pb2.ServerMessage.FromString,
            _registered_method=True,
        )


class FlexibleServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Send(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_FlexibleServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Send": grpc.stream_stream_rpc_method_handler(
            servicer.Send,
            request_deserializer=transport__pb2.ClientMessage.FromString,
            response_serializer=transport__pb2.ServerMessage.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "flexible.Flexible", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers("flexible.Flexible", rpc_method_handlers)


# This class is part of an EXPERIMENTAL API.
class Flexible(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Send(
        request_iterator,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            "/flexible.Flexible/Send",
            transport__pb2.ClientMessage.SerializeToString,
            transport__pb2.ServerMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import inspect
import os
import zipfile
from typing import Callable

import gdown
from sultan.api import Sultan
from tqdm import tqdm

EMNIST_DIGITS_URL = "https://drive.google.com/file/d/1fl9fRPPxTUxnC56ACzZ8JiLiew0SMFwt/view?usp=share_link"
EMNIST_DIGITS_MD5 = "5a18b33e88e3884e79f8b2d6274564d7"
EMNIST_DIGITS_FILE = "emnist-digits.mat"

EMNIST_LETTERS_URL = "https://drive.google.com/file/d/1KpwKUfx5L8zN0gPyrEuCEDFnC0OlEOoM/view?usp=share_link"
EMNIST_LETTERS_MD5 = "b9eddc3e325dee05b65fb21ee45da52f"
EMNIST_LETTERS_FILE = "emnist-letters.mat"

REDDIT_URL = (
    "https://drive.google.com/file/d/1S5R11eDR6g9Lkb02Ht6zmP2W-ZYqtS5U/view?usp=sharing"
)
REDDIT_MD5 = "308519a1945cf9268707d0aeba2b1c96"
REDDIT_FILE = "reddit_leaf.zip"

SHAKESPEARE_URL = "https://drive.google.com/file/d/1YkFqXqvN1s1JE1lEQ5YHb-aBPgmQ2IIk/view?usp=share_link"
SHAKESPEARE_MD5 = "a75e79f9ecebe5118314227a9cd19bc9"
SHAKESPEARE_FILE = "shakespeare_leaf.zip"


def check_integrity(filename: str, md5_hash: str) -> bool:
    """Function that computes the md5 hash of a file and compares it
        with a given one, ensuring that the file corresponds to the given md5 hash.

    Args:
    -----
        filename (str): path to file which will be used to compute a md5 hash
        md5_hash (str): md5 hash to compare with the one compute using filename

    Returns:
    --------
        bool: whether the given file has the same hash as the one provided
    """
    with Sultan.load() as s:
        try:
            result = s.md5("-q", filename).run()
        except Exception:
            result = s.md5sum(filename).pipe().cut("-f", "1", "-d", '" "').run()
    computed_md5 = result.stdout[0]
    return computed_md5 == md5_hash


def check_file_exists(filename: str) -> bool:
    """Function that checks if a given file exits or not.

    Args:
    -----
        filename (str): Filename to check.

    Returns:
    --------
        bool: True/False if the file exits or not.
    """
    return os.path.isfile(filename)


def check_dir_exists(filename: str) -> bool:
    """Function that checks if a given directory exits or not.

    Args:
    -----
        filename (str): Directory to check.

    Returns:
    --------
        bool: True/False if the directory exits or not.
    """
    return os.path.exists(filename)


def extract_zip(filename: str, output: bool = True):
    """Function that extract a zip file. If files are already extracted, it skips
    extracting them.

    Args:
    -----
        filename (str): Directory to check.
        output (bool): Whether to output the paths of the extracted files

    Returns:
    --------
        bool: True/False if the directory exits or not.
    """
    base_dir = "/".join(filename.split("/")[:-1])
    with zipfile.ZipFile(filename, "r", allowZip64=True) as zip_file:
        for member in zip_file.namelist():
            extracted_file_path = f"{base_dir}/{member}"
            if not (
                os.path.isfile(extracted_file_path)
                or os.path.exists(extracted_file_path)
            ):
                zip_file.extract(member, base_dir)
        if output:
            return [f"{base_dir}/{member}" for member in zip_file.namelist()]


def download_file(url: str, filename: str, out_dir: str = "."):
    """Function that downloads a file from a url and stores it in out_dir
        with name filename.

    Args:
    -----
        url (str): url to download the file
        filename (str): name used to store the downloaded file
        out_dir (str, optional): directory where the downloaded file will be stored. Defaults to ".".

    """
    additional_args = ()
    out_path = os.path.join(out_dir, filename)
    if "drive.google" not in url:
        additional_args = ("-#", "-L", "--output", out_path)
        with Sultan.load() as s:
            try:
                result = s.curl(url, *additional_args).run(streaming=True)
            except Exception:
                result = s.wget(url, "-O", out_path).run(streaming=True)

            def generator():
                while True:
                    complete = result.is_complete
                    if complete:
                        break
                    yield from result.stderr

            pbar = tqdm(generator())
            for i in pbar:
                pbar.set_description(i)
    else:  # Google drive download using gdown library
        gdown.download(url, quiet=False, fuzzy=True, output=out_path)


# Example: download_dataset(MNIST_URL, MNIST_FILE, MNIST_MD5, extract=True)
def download_dataset(
    url: str,
    filename: str,
    md5_hash: str,
    out_dir: str = ".",
    extract: bool = False,
    max_trials: int = 3,
    output: bool = True,
):
    """Function that download a dataset given an URL.

    Args:
    -----
        url (str): url to download the file
        filename (str): name used to store the downloaded file
        md5_hash (str): hash used to ensure the integrity of the downloaded file
        out_dir (str, optional): directory where the downloaded file will be stored. Defaults to ".".
        extract (bool, optional): Select to whether to extract the data or not. Defaults to False.
        max_trials (int, optional): Max number of trials to download the dataset. Defaults to 3.
        output (bool, optional): whether to return a list with the paths of the downloaded/extractred files. Defaults to True.

    Raises:
    -------
        ValueError: Raise an error if it fails downloading the dataset or the given md5 hash is not correct.
    """
    full_path = os.path.join(out_dir, filename)
    check_dir_exists(out_dir)
    i = 0
    while not (check_file_exists(full_path) and check_integrity(full_path, md5_hash)):
        download_file(url, filename, out_dir)
        i += 1
        if i > max_trials:
            raise ValueError(
                "Either we are unable the download the file or the provided md5 hash is not correct."
            )
    if extract:
        extracted_files = extract_zip(full_path, output=output)
    if output:
        return (
            [os.path.join(out_dir, i) for i in extracted_files]
            if extract
            else full_path
        )


def check_min_arguments(func: Callable, min_args: int = 1):
    """Function that inspect the minumum number of arguments of a given function.

    Args:
    -----
        func (Callable): Function to inspect
        min_args (int, optional): Minimum number of arguments that the function
        func must have. Defaults to 1.

    Raises:
    -------
        AssertionError: Raise an assertion error if the number of arguments of the
        given function is lower than the min_args value.
    """
    signature = inspect.signature(func)
    return len(signature.parameters) >= min_args



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import unittest

import pytest

from flex.actors.actors import FlexActors
from flex.actors.role import FlexRole
from flex.actors.role_manager import FlexRoleManager


@pytest.fixture(name="actors_cs")
def fixture_actors():
    client1_id = "client1"
    role1 = FlexRole.client
    client2_id = "client2"
    role2 = FlexRole.client
    client3_id = "client3"
    role3 = FlexRole.client
    aggregator_id = "aggregator"
    role4 = FlexRole.aggregator
    server_id = "server"
    role5 = FlexRole.server
    return FlexActors(
        {
            client1_id: role1,
            client2_id: role2,
            client3_id: role3,
            aggregator_id: role4,
            server_id: role5,
        }
    )


@pytest.fixture(name="p2p")
def fixture_peer_to_peer():
    actor1_id = "actor1"
    role1 = FlexRole.server_aggregator_client
    actor2_id = "actor2"
    role2 = FlexRole.server_aggregator_client
    actor3_id = "actor3"
    role3 = FlexRole.server_aggregator_client
    actor4_id = "actor4"
    role4 = FlexRole.server_aggregator_client
    actor5_id = "actor5"
    role5 = FlexRole.server_aggregator_client
    return FlexActors(
        {
            actor1_id: role1,
            actor2_id: role2,
            actor3_id: role3,
            actor4_id: role4,
            actor5_id: role5,
        }
    )


class TestRoleManger(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def _fixture_simple_actors_client_server(self, actors_cs):
        self._actors_cs = actors_cs

    @pytest.fixture(autouse=True)
    def _fixture_peer_to_peer(self, p2p):
        self._p2p = p2p

    def test_is_client_method(self):
        # Client-Server Architecture
        assert FlexRoleManager.is_client(self._actors_cs["client1"]) is True
        assert FlexRoleManager.is_client(self._actors_cs["client2"]) is True
        assert FlexRoleManager.is_client(self._actors_cs["client3"]) is True
        assert FlexRoleManager.is_client(self._actors_cs["aggregator"]) is False
        assert FlexRoleManager.is_client(self._actors_cs["server"]) is False
        # Peer-To-Peer Architecture
        assert FlexRoleManager.is_client(self._p2p["actor1"]) is True
        assert FlexRoleManager.is_client(self._p2p["actor2"]) is True
        assert FlexRoleManager.is_client(self._p2p["actor3"]) is True
        assert FlexRoleManager.is_client(self._p2p["actor4"]) is True
        assert FlexRoleManager.is_client(self._p2p["actor5"]) is True

    def test_is_aggregator_method(self):
        # Client-Server Architecture
        assert FlexRoleManager.is_aggregator(self._actors_cs["client1"]) is False
        assert FlexRoleManager.is_aggregator(self._actors_cs["client2"]) is False
        assert FlexRoleManager.is_aggregator(self._actors_cs["client3"]) is False
        assert FlexRoleManager.is_aggregator(self._actors_cs["aggregator"]) is True
        assert FlexRoleManager.is_aggregator(self._actors_cs["server"]) is False
        # Peer-To-Peer Architecture
        assert FlexRoleManager.is_aggregator(self._p2p["actor1"]) is True
        assert FlexRoleManager.is_aggregator(self._p2p["actor2"]) is True
        assert FlexRoleManager.is_aggregator(self._p2p["actor3"]) is True
        assert FlexRoleManager.is_aggregator(self._p2p["actor4"]) is True
        assert FlexRoleManager.is_aggregator(self._p2p["actor5"]) is True

    def test_is_server_method(self):
        # Client-Server Architecture
        assert FlexRoleManager.is_server(self._actors_cs["client1"]) is False
        assert FlexRoleManager.is_server(self._actors_cs["client2"]) is False
        assert FlexRoleManager.is_server(self._actors_cs["client3"]) is False
        assert FlexRoleManager.is_server(self._actors_cs["aggregator"]) is False
        assert FlexRoleManager.is_server(self._actors_cs["server"]) is True
        # Peer-To-Peer Architecture
        assert FlexRoleManager.is_server(self._p2p["actor1"]) is True
        assert FlexRoleManager.is_server(self._p2p["actor2"]) is True
        assert FlexRoleManager.is_server(self._p2p["actor3"]) is True
        assert FlexRoleManager.is_server(self._p2p["actor4"]) is True
        assert FlexRoleManager.is_server(self._p2p["actor5"]) is True

    def test_check_compatibility_method(self):
        # Client-Server Architecture
        assert (
            FlexRoleManager.check_compatibility(
                self._actors_cs["client1"], self._actors_cs["server"]
            )
            is False
        )
        assert (
            FlexRoleManager.check_compatibility(
                self._actors_cs["server"], self._actors_cs["client1"]
            )
            is True
        )
        assert (
            FlexRoleManager.check_compatibility(
                self._actors_cs["server"], self._actors_cs["aggregator"]
            )
            is False
        )
        assert (
            FlexRoleManager.check_compatibility(
                self._actors_cs["client1"], self._actors_cs["client2"]
            )
            is False
        )
        assert (
            FlexRoleManager.check_compatibility(
                self._actors_cs["aggregator"], self._actors_cs["server"]
            )
            is True
        )
        assert (
            FlexRoleManager.check_compatibility(
                self._actors_cs["client1"], self._actors_cs["aggregator"]
            )
            is True
        )
        # Peer-To-Peer Architecture
        assert (
            FlexRoleManager.check_compatibility(
                self._p2p["actor1"], self._p2p["actor2"]
            )
            is True
        )
        assert (
            FlexRoleManager.check_compatibility(
                self._p2p["actor1"], self._p2p["actor3"]
            )
            is True
        )
        assert (
            FlexRoleManager.check_compatibility(
                self._p2p["actor1"], self._p2p["actor4"]
            )
            is True
        )
        assert (
            FlexRoleManager.check_compatibility(
                self._p2p["actor1"], self._p2p["actor5"]
            )
            is True
        )
        assert (
            FlexRoleManager.check_compatibility(
                self._p2p["actor2"], self._p2p["actor3"]
            )
            is True
        )
        assert (
            FlexRoleManager.check_compatibility(
                self._p2p["actor2"], self._p2p["actor4"]
            )
            is True
        )
        assert (
            FlexRoleManager.check_compatibility(
                self._p2p["actor2"], self._p2p["actor5"]
            )
            is True
        )
        assert (
            FlexRoleManager.check_compatibility(
                self._p2p["actor3"], self._p2p["actor4"]
            )
            is True
        )
        assert (
            FlexRoleManager.check_compatibility(
                self._p2p["actor3"], self._p2p["actor5"]
            )
            is True
        )
        assert (
            FlexRoleManager.check_compatibility(
                self._p2p["actor4"], self._p2p["actor5"]
            )
            is True
        )
        assert (
            FlexRoleManager.check_compatibility(
                self._p2p["actor5"], self._p2p["actor3"]
            )
            is True
        )



import asyncio
import threading
import unittest

import numpy as np
import pytest

from flex.data.dataset import Dataset
from flex.distributed import ClientBuilder
from flex.distributed.aio.server import Server
from flex.model.model import FlexModel
from flex.pool.decorators import collect_clients_weights, set_aggregated_weights

addr, port = "localhost", 8080


@collect_clients_weights
def collect_weights(model):
    return [model["model"]["weights1"], model["model"]["weights2"]]


@set_aggregated_weights
def set_weights(model, weights):
    model["model"]["weights1"] = weights[0]
    model["model"]["weights2"] = weights[1]


def train(model, data):
    model["model"]["weights2"] = [7, 8, 9]
    return {"loss": 0.1}


def eval(model, data):
    return {"accuracy": 0.5}


def run_client():
    model = FlexModel()
    model["model"] = {"weights1": [1, 2, 3], "weights2": [4, 5, 6]}

    dataset = Dataset.from_array([[1, 2], [3, 4]], [0, 1])

    client = (
        ClientBuilder()
        .collect_weights(collect_weights)
        .model(model)
        .set_weights(set_weights)
        .train(train)
        .eval(eval, dataset)
        .dataset(dataset)
        .build()
    )
    client.run(f"{addr}:{port}")


class TestAsyncServer(unittest.IsolatedAsyncioTestCase):
    @pytest.mark.asyncio
    async def test_server_integration_test(self):
        server = Server()
        await server.run(addr, port)

        client = threading.Thread(target=run_client)
        client.start()
        await asyncio.wait_for(server.wait_for_clients(1), timeout=5)

        ids = server.get_ids()
        self.assertEqual(ids, ["0"], "Client ID is not correct")
        # Ping
        await asyncio.wait_for(server.ping(node_ids=ids), timeout=5)

        # Collect weights from the client
        weights = await server.collect_weights(node_ids=ids)
        self.assertTrue(
            all(
                [
                    np.array_equal(w, x)
                    for w, x in zip(
                        weights[0], [np.array([1, 2, 3]), np.array([4, 5, 6])]
                    )
                ]
            )
        )

        # set weights
        new_weights = [np.array([10, 11, 12]), np.array([13, 14, 15])]
        await server.send_weights(weights=new_weights, node_ids=ids)
        weights = await server.collect_weights(node_ids=ids)
        self.assertTrue(
            all([np.array_equal(w, x) for w, x in zip(weights[0], new_weights)])
        )

        # train
        metrics = await server.train(node_ids=ids)
        self.assertIsInstance(metrics, dict)
        self.assertTrue("0" in metrics)
        self.assertTrue("loss" in metrics["0"])
        self.assertAlmostEqual(metrics["0"]["loss"], 0.1)
        # Check that train has changed weights
        weights = await server.collect_weights(node_ids=ids)
        self.assertTrue(
            all(
                [
                    np.array_equal(w, x)
                    for w, x in zip(
                        weights[0], [np.array([10, 11, 12]), np.array([7, 8, 9])]
                    )
                ]
            ),
            f"weights: {weights}, expected: {[np.array([10, 11, 12]), np.array([7, 8, 9])]}",
        )

        # Eval
        metrics = await server.eval(node_ids=ids)
        self.assertIsInstance(metrics, dict)
        self.assertTrue("0" in metrics)
        self.assertTrue("accuracy" in metrics["0"])
        self.assertAlmostEqual(metrics["0"]["accuracy"], 0.5)

        await server.stop()
        client.join()



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import unittest

import pytest

from flex.actors.architectures import client_server_architecture, p2p_architecture
from flex.actors.role_manager import FlexRoleManager


@pytest.fixture(name="nodes_ids")
def fixture_nodes_ids():
    """Function that returns a Iterable provided as example to test functions.

    Returns
    -------
        List: A list with the node ids.
    """
    return ["client_1", "client_2", "client_3"]


class TestArchitectures(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def _fixture_nodes_ids(self, nodes_ids):
        self._nodes_ids = nodes_ids

    def test_client_server_architecture(self):
        architecture = client_server_architecture(self._nodes_ids)
        assert FlexRoleManager.is_client(architecture["client_1"]) is True
        assert FlexRoleManager.is_aggregator(architecture["client_1"]) is False
        assert FlexRoleManager.is_server(architecture["client_1"]) is False

        assert FlexRoleManager.is_client(architecture["client_2"]) is True
        assert FlexRoleManager.is_aggregator(architecture["client_2"]) is False
        assert FlexRoleManager.is_server(architecture["client_2"]) is False

        assert FlexRoleManager.is_client(architecture["client_3"]) is True
        assert FlexRoleManager.is_aggregator(architecture["client_3"]) is False
        assert FlexRoleManager.is_server(architecture["client_3"]) is False

        assert FlexRoleManager.is_client(architecture["server"]) is False
        assert FlexRoleManager.is_aggregator(architecture["server"]) is True
        assert FlexRoleManager.is_server(architecture["server"]) is True

    def test_p2p_architecture(self):
        architecture = p2p_architecture(self._nodes_ids)
        assert FlexRoleManager.is_client(architecture["client_1"]) is True
        assert FlexRoleManager.is_aggregator(architecture["client_1"]) is True
        assert FlexRoleManager.is_server(architecture["client_1"]) is True

        assert FlexRoleManager.is_client(architecture["client_2"]) is True
        assert FlexRoleManager.is_aggregator(architecture["client_2"]) is True
        assert FlexRoleManager.is_server(architecture["client_2"]) is True

        assert FlexRoleManager.is_client(architecture["client_3"]) is True
        assert FlexRoleManager.is_aggregator(architecture["client_3"]) is True
        assert FlexRoleManager.is_server(architecture["client_3"]) is True



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from flex.actors.role import FlexRole


class FlexRoleManager:
    """Class used to check allowed communications between
    different roles.
    """

    client_allowed_comm = {
        FlexRole.aggregator,
        FlexRole.aggregator_client,
        FlexRole.server_aggregator,
        FlexRole.server_aggregator_client,
    }
    aggregator_allowed_comm = {
        FlexRole.aggregator,
        FlexRole.aggregator_client,
        FlexRole.server,
        FlexRole.server_aggregator,
        FlexRole.server_client,
        FlexRole.server_aggregator_client,
    }
    server_allowed_comm = {
        FlexRole.client,
        FlexRole.aggregator_client,
        FlexRole.server_client,
        FlexRole.server_aggregator,
        FlexRole.server_aggregator_client,
    }

    client_wannabe = {
        FlexRole.client,
        FlexRole.server_client,
        FlexRole.aggregator_client,
        FlexRole.server_aggregator_client,
    }
    aggregator_wannabe = {
        FlexRole.aggregator,
        FlexRole.server_aggregator,
        FlexRole.aggregator_client,
        FlexRole.server_aggregator_client,
    }
    server_wannabe = {
        FlexRole.server,
        FlexRole.server_client,
        FlexRole.server_aggregator,
        FlexRole.server_aggregator_client,
    }

    @classmethod
    def is_client(cls, role: FlexRole) -> bool:
        """Method to check whether a role is a client role.

        Args:
        -----
            role (Role): role to be checked

        Returns:
        --------
            bool: wheter the not role is a client role
        """
        return role in cls.client_wannabe

    @classmethod
    def can_comm_with_client(cls, role: FlexRole) -> bool:
        """Method to ensure that role can establish a communication with
        a client role.

        Args:
        -----
            role (Role): role to be checked

        Returns:
        --------
            bool: whether or not role can communicate with a client role
        """
        return role in cls.client_allowed_comm

    @classmethod
    def is_aggregator(cls, role: FlexRole) -> bool:
        """Method to check whether a role is an aggregator role.

        Args:
        -----
            role (Role): role to be checked

        Returns:
        --------
            bool: wheter the not role is an aggregator role
        """
        return role in cls.aggregator_wannabe

    @classmethod
    def can_comm_with_aggregator(cls, role: FlexRole) -> bool:
        """Method to ensure that role can establish a communication with
        an aggregator role.

        Args:
        -----
            role (Role): role to be checked

        Returns:
        --------
            bool: whether or not role can communicate with a aggregator role
        """
        return role in cls.aggregator_allowed_comm

    @classmethod
    def is_server(cls, role: FlexRole) -> bool:
        """Method to check whether a role is a server role.

        Args:
        -----
            role (Role): role to be checked

        Returns:
        --------
            bool: wheter the not role is a server role
        """
        return role in cls.server_wannabe

    @classmethod
    def can_comm_with_server(cls, role: FlexRole) -> bool:
        """Method to ensure that role can establish a communication with
        a server role.

        Args:
        -----
            role (Role): role to be checked

        Returns:
        --------
            bool: whether or not role can communicate with a server role
        """
        return role in cls.server_allowed_comm

    @classmethod
    def check_compatibility(cls, role1: FlexRole, role2: FlexRole) -> bool:
        """Method used to ensure that it is possible to communicate from role1
        to role2, note that the communication from role2 to role1 is not checked.

        Args:
        -----
            role1 (Role): role which establishes communication with role2
            role2 (Role): role which receives communication from role1

        Returns:
        --------
            bool: whether or not the communication from role1 to role2 is allowed.
        """
        return any(
            [
                cls.is_client(role1) and cls.can_comm_with_client(role2),
                cls.is_aggregator(role1) and cls.can_comm_with_aggregator(role2),
                cls.is_server(role1) and cls.can_comm_with_server(role2),
            ]
        )



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from enum import Enum, unique


@unique
class FlexRole(Enum):
    """Enum which contains all possible roles:
        - Basic roles: client, server or aggregator
        - Composite roles: aggregator_client, server_client, server_aggregator, server_aggregator_client.

    Note that composite roles are designed to represented a combination of Basic roles.
    """

    client = 1
    aggregator = 2
    server = 3
    aggregator_client = 4
    server_client = 5
    server_aggregator = 6
    server_aggregator_client = 7



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import random
import unittest

import pytest
import tensorly as tl

from flex.pool.aggregators import fed_avg


def simulate_clients_weights_for_module(n_nodes, modulename):
    framework = __import__(modulename)
    n_nodes = 5
    simulated_client_weights = []
    num_layers = 5
    num_dim = [1, 2, 3, 4, 5]
    layer_ndims = random.sample(num_dim, k=num_layers)
    for _ in range(n_nodes):
        random.seed(0)
        simulated_weights = []
        for ndims in layer_ndims:
            tmp_dims = random.sample(
                num_dim, k=ndims
            )  # ndims dimensions of num_dim sizes
            simulated_weights.append(framework.ones(tmp_dims))
        simulated_client_weights.append(simulated_weights)
    return {"weights": simulated_client_weights}


class TestFlexAggregators(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def _fixture_weights(self):
        self._torch_weights = simulate_clients_weights_for_module(
            n_nodes=5, modulename="torch"
        )
        self._tf_weights = simulate_clients_weights_for_module(
            n_nodes=5, modulename="tensorflow"
        )
        self._np_weights = simulate_clients_weights_for_module(
            n_nodes=5, modulename="numpy"
        )

    def test_fed_avg_with_torch(self):
        fed_avg(self._torch_weights, None)
        agg_weights = self._torch_weights["aggregated_weights"]
        assert tl.get_backend() == "pytorch"
        assert all(
            tl.all(w == tl.ones(tl.shape(w), **tl.context(w))) for w in agg_weights
        )

    def test_fed_avg_with_tf(self):
        fed_avg(self._tf_weights, None)
        agg_weights = self._tf_weights["aggregated_weights"]
        assert tl.get_backend() == "tensorflow"
        assert all(
            tl.all(w == tl.ones(tl.shape(w), **tl.context(w))) for w in agg_weights
        )

    def test_fed_avg_with_np(self):
        fed_avg(self._np_weights, None)
        agg_weights = self._np_weights["aggregated_weights"]
        assert tl.get_backend() == "numpy"
        assert all(
            tl.all(w == tl.ones(tl.shape(w), **tl.context(w))) for w in agg_weights
        )



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
import signal
import sys
import threading
from abc import ABC, abstractmethod
from queue import Empty, Queue
from typing import List, Optional

import grpc
import numpy as np

from flex.data import Dataset
from flex.distributed.common import toNumpyArray, toTensorList
from flex.distributed.proto.transport_pb2 import ClientMessage, ServerMessage
from flex.distributed.proto.transport_pb2_grpc import FlexibleStub
from flex.model import FlexModel

logger = logging.getLogger(__name__)


class Client(ABC):
    """
    Allows the implementation for a client in a distributed `FLEXible` environment.
    The abstract methods must be implemented since they teach the client how to retrieve and set weights,
    run training and evaluation. The __init__ method must be called by the child class.
    """

    def __init__(self, dataset: Dataset, model: FlexModel, eval_dataset: Dataset):
        """
        Initializes a Client object.

        Args:
        ----
            dataset (Dataset): The dataset used for training.
            model (FlexModel): The model used for training.
            eval_dataset (Dataset): The dataset used for evaluation.
        """
        self._channel = None
        self._stub = None
        self._q = Queue()
        self._q.put(ClientMessage(handshake_res=ClientMessage.HandshakeRes(status=200)))
        self._finished = threading.Event()
        self.dataset = dataset
        self.model = model
        self.eval_dataset = eval_dataset

    def _iter_queue(self, q: Queue):
        # The iterator may still run if it has still messages to send
        while not self._finished.is_set() or not q.empty():
            try:
                value = q.get(timeout=0.1)
                yield value
            except Empty:
                # This condition is done in order to check finished
                continue
        raise StopIteration

    def _handle_get_weights_ins(self, response: ServerMessage.GetWeightsIns):
        logger.info("Weights requested")
        weights = self.get_weights(self.model)

        self._q.put(
            ClientMessage(
                get_weights_res=ClientMessage.GetWeightsRes(
                    weights=toTensorList([np.array(w) for w in weights])
                )
            )
        )
        logger.info("Weights sent")

    def _handle_send_weights_ins(self, response: ServerMessage.SendWeightsIns):
        logger.info("Weights received")
        weights = toNumpyArray(response.weights)
        self.set_weights(self.model, weights)
        self._q.put(
            ClientMessage(send_weights_res=ClientMessage.SendWeightsRes(status=200))
        )
        logger.info("Weights set")

    def _handle_health_ping(self, response: ServerMessage.HealthPing):
        logger.info("Health ping from server")
        self._q.put(ClientMessage(health_ins=ClientMessage.HealthPing(status=200)))

    @abstractmethod
    def set_weights(self, model: FlexModel, weights: List[np.ndarray]):
        """
        Sets the weights of the given model.

        Args:
        ----
            model (FlexModel): The model to set the weights for.
            weights (List[np.ndarray]): The weights to set for the model.
        """
        pass

    @abstractmethod
    def get_weights(self, model: FlexModel) -> List[np.ndarray]:
        """
        Retrieves the weights of a given FlexModel.

        Args:
        ----
            model (FlexModel): The FlexModel object for which to retrieve the weights.

        Returns:
        -------
            List[np.ndarray]: A list of NumPy arrays representing the weights of the model.
        """
        pass

    def _handle_train_ins(self, response: ServerMessage.TrainIns):
        logger.info("Training requested")
        metrics = self.train(self.model, self.dataset)
        if metrics is None or not isinstance(metrics, dict):
            metrics = {}

        self._q.put(ClientMessage(train_res=ClientMessage.TrainRes(metrics=metrics)))
        logger.info("Training completed")

    @abstractmethod
    def train(self, model: FlexModel, data: Dataset) -> dict:
        """
        Trains the given model using the provided dataset.

        Args:
        ----
            model (FlexModel): The model to be trained.
            data (Dataset): The dataset used for training.

        Returns:
        -------
            dict: A dictionary containing training metrics.
        """
        pass

    def _handle_eval_ins(self, response: ServerMessage.EvalIns):
        logger.info("Evaluation requested")
        metrics = self.eval(self.model, self.eval_dataset)
        if metrics is None or not isinstance(metrics, dict):
            metrics = {}

        self._q.put(ClientMessage(eval_res=ClientMessage.EvalRes(metrics=metrics)))
        logger.info("Evaluation completed")

    @abstractmethod
    def eval(self, model: FlexModel, data: Dataset) -> dict:
        """
        Evaluates the given model on the provided data.

        Args:
        ----
            model (FlexModel): The model to evaluate.
            data (Dataset): The data to evaluate the model on.

        Returns:
        -------
            dict: A dictionary containing evaluation metrics.
        """
        pass

    def run(
        self,
        address: str,
        root_certificate: str = None,
        _stub: Optional[FlexibleStub] = None,
    ):
        if _stub is not None:
            self._stub = _stub
        else:
            if root_certificate is not None:
                self._channel = grpc.secure_channel(
                    target=address,
                    credentials=grpc.ssl_channel_credentials(root_certificate),
                )
            else:
                self._channel = grpc.insecure_channel(address)

            self._stub = FlexibleStub(self._channel)

        self._request_generator = self._stub.Send(self._iter_queue(self._q))
        self._set_exit_handler(self._request_generator)
        try:
            for response in self._request_generator:
                assert isinstance(response, ServerMessage)
                msg = response.WhichOneof("msg")
                if msg == "get_weights_ins":
                    self._handle_get_weights_ins(response.get_weights_ins)
                elif msg == "send_weights_ins":
                    self._handle_send_weights_ins(response.send_weights_ins)
                elif msg == "train_ins":
                    self._handle_train_ins(response.train_ins)
                elif msg == "eval_ins":
                    self._handle_eval_ins(response.eval_ins)
                elif msg == "health_ins":
                    self._handle_health_ping(response.health_ins)
                else:
                    raise Exception("Not implemented")
        except Exception as e:
            cancelled = False
            if hasattr(e, "details") and callable(e.details):
                cancelled = "Cancelling all calls" in e.details()
            if cancelled:
                logger.info("Disconnected from server")
            else:
                logger.error(f"Canceling process. Got error: {e}")
            self._request_generator.cancel()
        finally:
            self._finished.set()

    def _set_exit_handler(self, request_generator):
        def _handler_(signum, frame):
            request_generator.cancel()
            sys.exit(0)

        # Only set the signal handler if the current thread is the main thread, avoids exceptions on running tests where
        # this code executes in another thread
        if threading.current_thread().__class__.__name__ == "_MainThread":
            signal.signal(signal.SIGINT, _handler_)

    def __del__(self):
        if self._channel is not None:
            self._channel.close()
